import scanpy as sc
import os
import torch
import numpy as np
from ..externel.gaston import neural_net,process_NN_output,dp_related,cluster_plotting
from scipy.sparse import issparse, csr_matrix
import matplotlib.pyplot as plt
from tqdm import tqdm
from .._settings import add_reference

class GASTON(object):

    def __init__(self,adata) -> None:
        self.adata=adata

    def get_gaston_input(self,get_rgb=False, spot_umi_threshold=50):
        import squidpy as sq
        adata=self.adata
        adata.obsm['spatial']=adata.obsm['spatial'].astype(float)
        sc.pp.filter_cells(adata, min_counts=spot_umi_threshold)

        gene_labels=adata.var.index.to_numpy()

        counts_mat=adata.X
        coords_mat=np.array(adata.obsm['spatial'])

        if not get_rgb:
            return counts_mat, coords_mat, gene_labels

        library_id = list(adata.uns['spatial'].keys())[0] # adata2.uns['spatial'] should have only one key
        scale=adata.uns['spatial'][library_id]['scalefactors']['tissue_hires_scalef']
        img = sq.im.ImageContainer(adata.uns['spatial'][library_id]['images']['hires'],
                                scale=scale,
                                layer="img1")
        print('calculating RGB')
        sq.im.calculate_image_features(adata, img, features="summary", key_added="features")
        columns = ['summary_ch-0_mean', 'summary_ch-1_mean', 'summary_ch-2_mean']
        RGB_mean = adata.obsm["features"][columns]
        RGB_mean = RGB_mean / 255
        self.RGB_mean=RGB_mean
        
        # RGB_mean = RGB_mean[RGB_mean.index.isin(df_pos['barcode'])]

        adata=self.adata
        return counts_mat, coords_mat, gene_labels, RGB_mean.to_numpy()
    
    def get_top_pearson_residuals(self,num_dims=5,clip=0.01,n_top_genes=5000,
                                  use_RGB=False):
        adata=self.adata
        sc.experimental.pp.highly_variable_genes(
            adata, flavor="pearson_residuals", n_top_genes=n_top_genes
        )


        adata = adata[:, adata.var["highly_variable"]]
        adata.layers["raw"] = adata.X.copy()
        adata.layers["sqrt_norm"] = np.sqrt(
            sc.pp.normalize_total(adata, inplace=False)["X"]
        )

        theta=np.Inf
        sc.experimental.pp.normalize_pearson_residuals(adata, clip=clip, theta=theta)
        sc.pp.pca(adata, n_comps=num_dims)
        A=adata.obsm['X_pca']
        if use_RGB:
            A=np.hstack((A,self.RGB_mean)) # attach to RGB mean
        return A
    
    def load_rescale(self,A):
        S=self.adata.obsm['spatial']
        from ..externel.gaston.neural_net import load_rescale_input_data
        S_torch, A_torch = load_rescale_input_data(S,A)
        self.S_torch=S_torch
        self.A_torch=A_torch
        #return S_torch, A_torch
    
    def train(self,isodepth_arch=[20,20],expression_fn_arch=[20,20],
              num_epochs=10000,checkpoint=500,out_dir='result/test_outputs',
              optimizer="adam",num_restarts=30):
        ######################################
        # NEURAL NET PARAMETERS (USER CAN CHANGE)
        # architectures are encoded as list, eg [20,20] means two hidden layers of size 20 hidden neurons
        #isodepth_arch=[20,20] # architecture for isodepth neural network d(x,y) : R^2 -> R 
        #expression_fn_arch=[20,20] # architecture for 1-D expression function h(w) : R -> R^G

       # num_epochs = 10000 # number of epochs to train NN (NOTE: it is sometimes beneficial to train longer)
        #checkpoint = 500 # save model after number of epochs = multiple of checkpoint
        #out_dir='colorectal_tumor_tutorial_outputs' # folder to save model runs
        #optimizer = "adam"
        #num_restarts=30

        ######################################
        #mkdir out_dir
        os.makedirs(out_dir, exist_ok=True)

        seed_list=range(num_restarts)
        for seed in tqdm(seed_list):
            #print(f'training neural network for seed {seed}')
            out_dir_seed=f"{out_dir}/rep{seed}"
            os.makedirs(out_dir_seed, exist_ok=True)
            mod, loss_list = neural_net.train(self.S_torch, self.A_torch,
                                S_hidden_list=isodepth_arch, A_hidden_list=expression_fn_arch, 
                                epochs=num_epochs, checkpoint=checkpoint, 
                                save_dir=out_dir_seed, optim=optimizer, seed=seed, save_final=True)
            
        add_reference(self.adata,'GASTON','spatial depth estimation with GASTON')

    def get_best_model(self,out_dir='result/test_outputs',
                       max_domain_num=8,start_from=2):
        gaston_model, A, S= process_NN_output.process_files(out_dir) # MATCH PAPER FIGURES
        self.model=gaston_model
        self.A=A
        self.S=S
        from ..externel.gaston import model_selection
        model_selection.plot_ll_curve(gaston_model, A, S, max_domain_num=max_domain_num, start_from=start_from)
        return gaston_model, A, S
    
    def cal_iso_depth(self,num_domains=10):
        #num_domains=5 # CHANGE FOR YOUR APPLICATION: use number of layers from above!
        gaston_isodepth, gaston_labels=dp_related.get_isodepth_labels(self.model,
                                                                      self.A,
                                                                      self.S,
                                                                      num_domains)


        # DATASET-SPECIFIC: so domains are ordered with tumor being last
        gaston_isodepth= np.max(gaston_isodepth) -1 * gaston_isodepth
        gaston_labels=(num_domains-1)-gaston_labels
        self.gaston_isodepth=gaston_isodepth
        self.gaston_labels=gaston_labels
        return gaston_isodepth, gaston_labels
    
    def plot_isodepth(self,show_streamlines=True,
                      rotate_angle=-90,arrowsize=2,
                      figsize=(7,6),**kwargs):
        
        rotate = np.radians(rotate_angle) # rotate coordinates by -90

        cluster_plotting.plot_isodepth(self.gaston_isodepth, 
                                       self.S, 
                                       self.model, figsize=figsize, streamlines=show_streamlines, 
                                    rotate=rotate,arrowsize=arrowsize, 
                                    **kwargs) # since we did isodepth -> -1*isodepth above, we also need to do gradient -> -1*gradient

    def plot_clusters(self,domain_colors,figsize=(6,6),
                      s=20,lgd=False,show_boundary=True,
                      rotate_angle=-90,boundary_lw=5,**kwargs):
        rotate = np.radians(rotate_angle) # rotate coordinates by -90
        cluster_plotting.plot_clusters(self.gaston_labels, self.S, figsize=figsize, 
                               colors=domain_colors, s=s, lgd=lgd, 
                               show_boundary=show_boundary, 
                               gaston_isodepth=self.gaston_isodepth, 
                               boundary_lw=boundary_lw, rotate=rotate,
                               **kwargs)
        
    def plot_clusters_restrict(self,domain_colors,isodepth_min=4.5,isodepth_max=6.8,
                               rotate_angle=-90,s=20,lgd=False,figsize=(6,6), **kwargs):
        # This is the range we used for reproducing figure papers. We found these bounds manually.
        rotate = np.radians(rotate_angle) # rotate coordinates by -90
        cluster_plotting.plot_clusters_restrict(self.gaston_labels, self.S, self.gaston_isodepth, 
                                                isodepth_min=isodepth_min, isodepth_max=isodepth_max, figsize=figsize, 
                                                colors=domain_colors, s=s, lgd=lgd, rotate=rotate, **kwargs)
        
    def restrict_spot(self,isodepth_min=4.5,isodepth_max=6.8,
                      adjust_physical=True,scale_factor=100,
                      plotisodepth=True,show_streamlines=True,
                      rotate_angle=-90,arrowsize=1, figsize=(6,3), 
                      neg_gradient=True,
                      **kwargs):
        # Optional: adjust isodepth for physical distance
        rotate = np.radians(rotate_angle) # rotate coordinates by -90

        from ..externel.gaston.restrict_spots import restrict_spots
        counts_mat_restrict, coords_mat_restrict, gaston_isodepth_restrict, gaston_labels_restrict, S_restrict=restrict_spots(
                                                                    self.adata.X, 
                                                                    self.adata.obsm['spatial'], 
                                                                    self.S, self.gaston_isodepth, self.gaston_labels, 
                                                                    isodepth_min=isodepth_min, isodepth_max=isodepth_max, 
                                                                    adjust_physical=adjust_physical, scale_factor=scale_factor,
                                                                    plotisodepth=plotisodepth, show_streamlines=show_streamlines, 
                                                                    gaston_model=self.model, rotate=rotate, figsize=figsize, 
                                                                    arrowsize=arrowsize, 
                                                                    neg_gradient=neg_gradient,**kwargs) # since we reversed gradient direction earlier, we need to reverse it back
        self.counts_mat_restrict=counts_mat_restrict
        self.coords_mat_restrict=coords_mat_restrict
        self.gaston_isodepth_restrict=gaston_isodepth_restrict
        self.gaston_labels_restrict=gaston_labels_restrict
        self.S_restrict=S_restrict

        return counts_mat_restrict, coords_mat_restrict, gaston_isodepth_restrict, gaston_labels_restrict, S_restrict
    
    def filter_genes(self,umi_thresh = 1000,exclude_prefix=['Mt-', 'Rpl', 'Rps']):

        #umi_thresh = 1000 # only analyze genes with at least 1000 total UMIs
        self.umi_thresh=umi_thresh
        counts_mat=self.adata.X
        if issparse(counts_mat):
            counts_mat=counts_mat.toarray()
        from ..externel.gaston.filter_genes import filter_genes
        idx_kept, gene_labels_idx=filter_genes(counts_mat, self.adata.var.index.to_numpy(), 
                                       umi_threshold=umi_thresh, 
                                       exclude_prefix=exclude_prefix)
        self.idx_kept=idx_kept
        self.gene_labels_idx=gene_labels_idx

    def pw_linear_fit(self,cell_type_df=None,
                        ct_list=[],**kwargs):
        from ..externel.gaston.segmented_fit import pw_linear_fit
        pw_fit_dict=pw_linear_fit(self.counts_mat_restrict, 
                                  self.gaston_labels_restrict, 
                                  self.gaston_isodepth_restrict,
                                        cell_type_df, ct_list,  idx_kept=self.idx_kept, 
                                        umi_threshold=self.umi_thresh, isodepth_mult_factor=0.01,
                                        **kwargs)
        self.pw_fit_dict=pw_fit_dict
        return pw_fit_dict
    
    def bin_data(self,cell_type_df=None,
                 num_bins=15,q_discont=0.95,q_cont=0.8,**kwargs):
        # for plotting, 
        from ..externel.gaston.binning_and_plotting import bin_data
        from ..externel.gaston.spatial_gene_classification import get_discont_genes, get_cont_genes
        binning_output=bin_data(self.counts_mat_restrict, 
                                self.gaston_labels_restrict, 
                                self.gaston_isodepth_restrict, 
                                cell_type_df, self.adata.var.index.to_numpy(), 
                                idx_kept=self.idx_kept, num_bins=num_bins, umi_threshold=self.umi_thresh,
                                **kwargs)
        self.binning_output=binning_output
        self.discont_genes_layer=get_discont_genes(self.pw_fit_dict, binning_output,q=q_discont)
        self.cont_genes_layer=get_cont_genes(self.pw_fit_dict, binning_output,q=q_cont)  
        return binning_output
    
    def get_restricted_adata(self,offset=10**6,):
        adata=self.adata
        adata2=adata[:,self.adata.var_names[self.idx_kept]]
        #adata2.obsm['spatial']=self.coords_mat_restrict
        adata2=adata2[:,self.gene_labels_idx]
        adata2.uns['gaston']={}
        adata2.uns['gaston']['isodepth']=self.gaston_isodepth_restrict
        adata2.uns['gaston']['labels']=self.gaston_labels_restrict

        slope_mat, intercept_mat, _, _ = self.pw_fit_dict['all_cell_types']

        gene_list = list(self.binning_output['gene_labels_idx']) # 获取基因列表
        adata2=adata2[:,gene_list]
        all_gene_outputs = []

        for gene_name in tqdm(gene_list):
            if gene_name in self.binning_output['gene_labels_idx']:
                gene_index = np.where(self.gene_labels_idx == gene_name)[0]

                outputs = np.zeros(self.gaston_isodepth_restrict.shape[0])
                for i in range(self.gaston_isodepth_restrict.shape[0]):
                    dom = int(self.gaston_labels_restrict[i])
                    slope = slope_mat[gene_index, dom]
                    intercept = intercept_mat[gene_index, dom]
                    outputs[i] = np.log(offset) + intercept + slope * self.gaston_isodepth_restrict[i]

                all_gene_outputs.append(outputs)

        sparse_output_matrix = csr_matrix(all_gene_outputs)
        adata2.layers['GASTON_ReX']=sparse_output_matrix.T
        return adata2
    
    def plot_gene_pwlinear(self,gene,domain_colors,offset=10**6,
                           cell_type_list=None,pt_size=50,linear_fit=True,
                           ticksize=15, figsize=(4,2.5),
                           lw=3,domain_boundary_plotting=True):
        gene_name=gene
        print(f'gene {gene_name}: discontinuous jump after domain(s) {self.discont_genes_layer[gene_name]}') 
        print(f'gene {gene_name}: continuous gradient in domain(s) {self.cont_genes_layer[gene_name]}')

        # display log CPM (if you want to do CP500, set offset=500)
        #offset=10**6
        from ..externel.gaston.binning_and_plotting import plot_gene_pwlinear
        plot_gene_pwlinear(gene_name, self.pw_fit_dict, 
                           self.gaston_labels_restrict, 
                           self.gaston_isodepth_restrict, 
                           self.binning_output, cell_type_list=cell_type_list, pt_size=pt_size, colors=domain_colors, 
                                                linear_fit=linear_fit, ticksize=ticksize, figsize=figsize, offset=offset, lw=lw,
                                            domain_boundary_plotting=domain_boundary_plotting)
        
    def plot_gene_raw(self,gene_name,rotate_angle=-90,
                      vmin=5,figsize=(6,3),s=10,**kwargs):
        rotate=np.radians(rotate_angle)
        from ..externel.gaston.binning_and_plotting import plot_gene_raw
        if issparse(self.counts_mat_restrict):
            counts_mat_restrict=self.counts_mat_restrict.toarray()
        else:
            counts_mat_restrict=self.counts_mat_restrict
        plot_gene_raw(gene_name, self.gene_labels_idx, counts_mat_restrict[:,self.idx_kept], 
                      self.S_restrict, vmin=vmin, figsize=figsize,s=s,rotate=rotate,**kwargs)
        plt.title(f'{gene_name} Raw Expression')

    def plot_gene_gastonrex(self,gene_name,rotate_angle=-90,
                            figsize=(6,3),s=10,**kwargs):
        rotate=np.radians(rotate_angle)
        from ..externel.gaston.binning_and_plotting import plot_gene_function

        plot_gene_function(gene_name, self.S_restrict, self.pw_fit_dict, 
                           self.gaston_labels_restrict, self.gaston_isodepth_restrict, 
                                        self.binning_output, figsize=figsize, s=s, rotate=rotate, **kwargs)
        plt.title(f'{gene_name} GASTON ReX')



from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union
import math
import os
import torch
import torch.nn as nn
from torch.nn import functional as F
from transformers.generation.utils import SampleDecoderOnlyOutput


class LayerNorm(nn.Module):

    def __init__(self, ndim, bias):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(ndim))
        self.bias = nn.Parameter(torch.zeros(ndim)) if bias else None

    def forward(self, input):
        return F.layer_norm(input, self.weight.shape, self.weight, self.bias, 1e-5)


class CausalSelfAttention(nn.Module):

    def __init__(self, config):
        super().__init__()
        assert config.n_embd % config.n_head == 0
        self.c_attn = nn.Linear(config.n_embd, 3 * config.n_embd, bias=config.bias)
        self.c_proj = nn.Linear(config.n_embd, config.n_embd, bias=config.bias)
        self.attn_dropout = nn.Dropout(config.dropout)
        self.resid_dropout = nn.Dropout(config.dropout)
        self.n_head = config.n_head
        self.n_embd = config.n_embd
        self.dropout = config.dropout
        self.config = config

    def forward(self, x, layer_past=None):
        q, k, v  = self.c_attn(x).split(self.n_embd, dim=2)
        B, T, C = k.size()
        _, Tq, _ = q.size()
        k = k.view(B, T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, T, hs)
        q = q.view(B, T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, T, hs)
        v = v.view(B, T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, T, hs)
   
        if layer_past is not None:
            past_key, past_value = layer_past
            k = torch.cat((past_key, k), dim=2)
            v = torch.cat((past_value, v), dim=2)
        B, nh, T, hs = k.size()
        
        if Tq < T:
            y = torch.nn.functional.scaled_dot_product_attention(q, k, v, attn_mask=None, dropout_p=self.dropout if self.training else 0, is_causal=False)
        else:
            y = torch.nn.functional.scaled_dot_product_attention(q, k, v, attn_mask=None, dropout_p=self.dropout if self.training else 0, is_causal=True)
        y = y.transpose(1, 2).contiguous().view(B, Tq, C) # re-assemble all head outputs side by side
    
        y = self.resid_dropout(self.c_proj(y))
        return y, (k, v)

def new_gelu(x):
    return 0.5 * x * (1.0 + torch.tanh(math.sqrt(2.0 / math.pi) * (x + 0.044715 * torch.pow(x, 3.0))))

class MLP(nn.Module):

    def __init__(self, config):
        super().__init__()
        self.c_fc    = nn.Linear(config.n_embd, 4 * config.n_embd, bias=config.bias)
        self.c_proj  = nn.Linear(4 * config.n_embd, config.n_embd, bias=config.bias)
        self.dropout = nn.Dropout(config.dropout)

    def forward(self, x):
        x = self.c_fc(x)
        x = new_gelu(x)
        x = self.c_proj(x)
        x = self.dropout(x)
        return x


class Block(nn.Module):

    def __init__(self, config):
        super().__init__()
        self.ln_1 = LayerNorm(config.n_embd, bias=config.bias)
        self.attn = CausalSelfAttention(config)
        self.ln_2 = LayerNorm(config.n_embd, bias=config.bias)
        self.mlp = MLP(config)

    def forward(self, x, layer_past=None):
        a,kv_cache = self.attn(self.ln_1(x),layer_past)
        x = x+a
        x = x + self.mlp(self.ln_2(x))
        return x,kv_cache

@dataclass
class MulanConfig:
    block_size: int = 1000
    vocab_size: int = 1011 
    n_layer: int = 12
    n_head: int = 12
    n_embd: int = 768
    dropout: float = 0.0
    bias: bool = True 
    train_mode: str = 'pretrian'
    expression_level: int = 10
    ele: int = 0


class scMulanModel_kv(nn.Module):
    def __init__(self, config):
        super().__init__()
        assert config.vocab_size is not None
        assert config.block_size is not None
        assert config.expression_level is not None
        assert config.ele == 1
        self.config = config

        self.transformer = nn.ModuleDict(dict(
            wte = nn.Embedding(config.vocab_size, config.n_embd),
            wee = nn.Embedding(config.expression_level + 1, config.n_embd), # +1 for non gene tokens
            drop = nn.Dropout(config.dropout),
            h = nn.ModuleList([Block(config) for _ in range(config.n_layer)]),
            ln_f = LayerNorm(config.n_embd, bias=config.bias),
        ))
            
        self.lm_head = nn.Linear(config.n_embd, config.vocab_size, bias=False)
        self.epx_head = nn.Linear(config.n_embd, config.vocab_size, bias=False) # expr level

        if 'LOCAL_RANK' not in os.environ or os.environ['LOCAL_RANK'] == '0':
            print("number of parameters: %.2fM" % (self.get_num_params()/1e6,))

        
    def get_num_params(self): 
        n_params = sum(p.numel() for p in self.parameters())

        return n_params
    
    def forward(self, 
                idx = None,
                inputs_embeds = None,
                past_key_values: Optional[Tuple[Tuple[torch.Tensor]]] = None,
                targets = None, 
                xlen = None,
                x_prefix_len = None,
                x_expr = None,
                y_expr = None,
                return_hidden = False,
                ):
        if past_key_values is None:
            past_length = 0
            past_key_values = tuple([None] * len(self.transformer.h))
        else:
            past_length = past_key_values[0][0].size(-2)
        # kvcache now
        presents_kv = () 
        if idx is not None:
            b, t = idx.size()
            tok_emb = self.transformer.wte(idx) # token embeddings of shape (b, t, n_embd)

        if inputs_embeds is not None:
            tok_emb = inputs_embeds
        expr_emb = self.transformer.wee(x_expr) # expression embeddings of shape (b, t, n_embd)
        x = self.transformer.drop(tok_emb + expr_emb)
        for i,(block, layer_past) in enumerate(zip(self.transformer.h, past_key_values)):
            x,layer_present = block(x,layer_past)
            presents_kv += (layer_present,)
        x = self.transformer.ln_f(x)

        logits_cls = self.lm_head(x[:, [-1], :]) # note: using list [-1] to preserve the time dim
        logits_exp = self.epx_head(x[:, [-1], :])
        loss = None
        loss_cls = None
        loss_exp = None
        if return_hidden:
            return logits_cls,logits_exp, x, presents_kv
        return logits_cls, logits_exp, loss, loss_cls, loss_exp, presents_kv

    @torch.no_grad()
    def generate_cellGenesis(self, 
                      input_ids,
                      expression_level,
                      max_new_tokens, 
                      ignore_Idx = None, 
                      top_k = None, 
                      return_dict_in_generate = False,
                      return_hidden = False,
                      gamma = 1):

        scores = ()
        hidden_states = ()
        past_key_values = None
        idx_cond = input_ids
        output_idx = idx_cond
        
        idx_expr = expression_level # 输入的token长度，一开始是3，后面变成nexttoken_expr 1。
        output_expr = expression_level # 输出的 token长度，一开始是3，后面变成nexttoken_expr 1。

        while True:
            if return_hidden:
                logits_cls,logits_exp,hidden,past_key_values = self(idx = idx_cond, x_expr = idx_expr, return_hidden = True, past_key_values = past_key_values)
                hidden_states += (hidden,)
            else:
                logits_cls, logits_exp, loss, loss_cls, loss_exp, past_key_values = self(idx = idx_cond, x_expr = idx_expr,past_key_values = past_key_values)
            logits_cls = logits_cls[:,-1,:] # (B,C)
            logits_exp = logits_exp[:,-1,:] # (B,C)

            if ignore_Idx is not None:
                # return logits, ignore_Idx
                logits_cls[:,ignore_Idx] = float('-inf')
            logits_cls[:,output_idx[0,:]] = float('-inf')
            
            if top_k is not None:
                v, _ = torch.topk(logits_cls, min(top_k, logits_cls.size(-1)))
                logits_cls[logits_cls < v[:, [-1]]] = float('-inf')
            
            next_token_scores = logits_cls

            if return_dict_in_generate:
                scores += (next_token_scores,)
            
            probs = F.softmax(logits_cls, dim=-1) #(B,C)
            probs[:,0] = gamma*probs[:,0]
           
            next_tokens = torch.multinomial(probs,num_samples=1) #(B,1)
            next_token_ele = logits_exp[torch.arange(logits_exp.size(0)),next_tokens.squeeze(1)].unsqueeze(1) # (B,1)
            bin_ele_next_token = torch.clamp(torch.round(next_token_ele), 0, 10).int()
            output_idx = torch.cat((output_idx,next_tokens),dim=1)
            output_expr = torch.cat((output_expr,bin_ele_next_token),dim=1)
            idx_cond = next_tokens
            idx_expr = bin_ele_next_token
            
            if next_tokens == 0 or len(output_idx[0]) >= max_new_tokens:
                break
            
        if return_dict_in_generate:
            return SampleDecoderOutput(
                sequences=output_idx,
                scores=scores,
                hidden_states=hidden_states,
                expression=output_expr,
                )
        elif return_hidden:
            return output_idx, output_expr, hidden_states
        else:return output_idx, output_expr

@dataclass
class SampleDecoderOutput(SampleDecoderOnlyOutput):

    sequences: torch.LongTensor = None
    scores: Optional[Tuple[torch.FloatTensor]] = None
    attentions: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    hidden_states: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    expression: Optional[torch.LongTensor] = None


"""Plotting utilities."""
from ._palette import *
from ._single import *
from ._general import *
from ._heatmap import *
from ._multi import *
from ._bulk import *
from ._space import *
from ._cpdb import *
from ._flowsig import *
from ._embedding import *
from ._density import *





import pandas as pd
import numpy as np
import sklearn.neighbors
import networkx as nx
from .mnn_utils import create_dictionary_mnn

def match_cluster_labels(true_labels,est_labels):
    true_labels_arr = np.array(list(true_labels))
    est_labels_arr = np.array(list(est_labels))
    org_cat = list(np.sort(list(pd.unique(true_labels))))
    est_cat = list(np.sort(list(pd.unique(est_labels))))
    B = nx.Graph()
    B.add_nodes_from([i+1 for i in range(len(org_cat))], bipartite=0)
    B.add_nodes_from([-j-1 for j in range(len(est_cat))], bipartite=1)
    for i in range(len(org_cat)):
        for j in range(len(est_cat)):
            weight = np.sum((true_labels_arr==org_cat[i])* (est_labels_arr==est_cat[j]))
            B.add_edge(i+1,-j-1, weight=-weight)
    match = nx.algorithms.bipartite.matching.minimum_weight_full_matching(B)
#     match = minimum_weight_full_matching(B)
    if len(org_cat)>=len(est_cat):
        return np.array([match[-est_cat.index(c)-1]-1 for c in est_labels_arr])
    else:
        unmatched = [c for c in est_cat if not (-est_cat.index(c)-1) in match.keys()]
        l = []
        for c in est_labels_arr:
            if (-est_cat.index(c)-1) in match: 
                l.append(match[-est_cat.index(c)-1]-1)
            else:
                l.append(len(org_cat)+unmatched.index(c))
        return np.array(l)      


def Cal_Spatial_Net(adata, rad_cutoff=None, k_cutoff=None,
                    max_neigh=50, model='Radius', verbose=True):
    """\
    Construct the spatial neighbor networks.

    Parameters
    ----------
    adata
        AnnData object of scanpy package.
    rad_cutoff
        radius cutoff when model='Radius'
    k_cutoff
        The number of nearest neighbors when model='KNN'
    model
        The network construction model. When model=='Radius', the spot is connected to spots whose distance is less than rad_cutoff. When model=='KNN', the spot is connected to its first k_cutoff nearest neighbors.

    Returns
    -------
    The spatial networks are saved in adata.uns['Spatial_Net']
    """

    assert (model in ['Radius', 'KNN'])
    if verbose:
        print('------Calculating spatial graph...')
    coor = pd.DataFrame(adata.obsm['spatial'])
    coor.index = adata.obs.index
    coor.columns = ['imagerow', 'imagecol']

    nbrs = sklearn.neighbors.NearestNeighbors(
        n_neighbors=max_neigh + 1, algorithm='ball_tree').fit(coor)
    distances, indices = nbrs.kneighbors(coor)
    if model == 'KNN':
        indices = indices[:, 1:k_cutoff + 1]
        distances = distances[:, 1:k_cutoff + 1]
    if model == 'Radius':
        indices = indices[:, 1:]
        distances = distances[:, 1:]

    KNN_list = []
    for it in range(indices.shape[0]):
        KNN_list.append(pd.DataFrame(zip([it] * indices.shape[1], indices[it, :], distances[it, :])))
    KNN_df = pd.concat(KNN_list)
    KNN_df.columns = ['Cell1', 'Cell2', 'Distance']

    Spatial_Net = KNN_df.copy()
    if model == 'Radius':
        Spatial_Net = KNN_df.loc[KNN_df['Distance'] < rad_cutoff,]
    id_cell_trans = dict(zip(range(coor.shape[0]), np.array(coor.index), ))
    Spatial_Net['Cell1'] = Spatial_Net['Cell1'].map(id_cell_trans)
    Spatial_Net['Cell2'] = Spatial_Net['Cell2'].map(id_cell_trans)
    # self_loops = pd.DataFrame(zip(Spatial_Net['Cell1'].unique(), Spatial_Net['Cell1'].unique(),
    #                  [0] * len((Spatial_Net['Cell1'].unique())))) ###add self loops
    # self_loops.columns = ['Cell1', 'Cell2', 'Distance']
    # Spatial_Net = pd.concat([Spatial_Net, self_loops], axis=0)

    if verbose:
        print('The graph contains %d edges, %d cells.' % (Spatial_Net.shape[0], adata.n_obs))
        print('%.4f neighbors per cell on average.' % (Spatial_Net.shape[0] / adata.n_obs))
    adata.uns['Spatial_Net'] = Spatial_Net

    #########
    X = pd.DataFrame(adata.X.toarray()[:, ], index=adata.obs.index, columns=adata.var.index)
    cells = np.array(X.index)
    cells_id_tran = dict(zip(cells, range(cells.shape[0])))
    if 'Spatial_Net' not in adata.uns.keys():
        raise ValueError("Spatial_Net is not existed! Run Cal_Spatial_Net first!")
        
    Spatial_Net = adata.uns['Spatial_Net']
    G_df = Spatial_Net.copy()
    G_df['Cell1'] = G_df['Cell1'].map(cells_id_tran)
    G_df['Cell2'] = G_df['Cell2'].map(cells_id_tran)
    G = sp.coo_matrix((np.ones(G_df.shape[0]), (G_df['Cell1'], G_df['Cell2'])), shape=(adata.n_obs, adata.n_obs))
    G = G + sp.eye(G.shape[0])  # self-loop
    adata.uns['adj'] = G
    

def Stats_Spatial_Net(adata):
    import matplotlib.pyplot as plt
    Num_edge = adata.uns['Spatial_Net']['Cell1'].shape[0]
    Mean_edge = Num_edge / adata.shape[0]
    plot_df = pd.value_counts(pd.value_counts(adata.uns['Spatial_Net']['Cell1']))
    plot_df = plot_df / adata.shape[0]
    fig, ax = plt.subplots(figsize=[3, 2])
    plt.ylabel('Percentage')
    plt.xlabel('')
    plt.title('Number of Neighbors (Mean=%.2f)' % Mean_edge)
    ax.bar(plot_df.index, plot_df)
    plt.show()


def mclust_R(adata, num_cluster, modelNames='EEE', used_obsm='STAGATE', random_seed=666):
    """\
    Clustering using the mclust algorithm.
    The parameters are the same as those in the R package mclust.
    """

    np.random.seed(random_seed)
    import rpy2.robjects as robjects
    robjects.r.library("mclust")

    import rpy2.robjects.numpy2ri
    rpy2.robjects.numpy2ri.activate()
    r_random_seed = robjects.r['set.seed']
    r_random_seed(random_seed)
    rmclust = robjects.r['Mclust']

    res = rmclust(adata.obsm[used_obsm], num_cluster, modelNames)
    mclust_res = np.array(res[-2])

    adata.obs['mclust'] = mclust_res
    adata.obs['mclust'] = adata.obs['mclust'].astype('int')
    adata.obs['mclust'] = adata.obs['mclust'].astype('category')
    return adata

import scipy.sparse as sp
def prepare_graph_data(adj):
    # adapted from preprocess_adj_bias
    num_nodes = adj.shape[0]
    # adj = adj + sp.eye(num_nodes)# self-loop  ##new !!
    #data =  adj.tocoo().data
    #adj[adj > 0.0] = 1.0
    if not sp.isspmatrix_coo(adj):
        adj = adj.tocoo()
    adj = adj.astype(np.float32)
    indices = np.vstack((adj.col, adj.row)).transpose()

    # adj = normalize(adj, norm="l1")

    return (adj, indices, adj.data, adj.shape)


def prune_spatial_Net(Graph_df, label):
    print('------Pruning the graph...')
    print('%d edges before pruning.' %Graph_df.shape[0])
    pro_labels_dict = dict(zip(list(label.index), label))
    Graph_df['Cell1_label'] = Graph_df['Cell1'].map(pro_labels_dict)
    Graph_df['Cell2_label'] = Graph_df['Cell2'].map(pro_labels_dict)
    Graph_df = Graph_df.loc[Graph_df['Cell1_label']==Graph_df['Cell2_label'],]
    print('%d edges after pruning.' %Graph_df.shape[0])
    return Graph_df


# https://github.com/ClayFlannigan/icp
def best_fit_transform(A, B):
    '''
    Calculates the least-squares best-fit transform that maps corresponding points A to B in m spatial dimensions
    Input:
      A: Nxm numpy array of corresponding points
      B: Nxm numpy array of corresponding points
    Returns:
      T: (m+1)x(m+1) homogeneous transformation matrix that maps A on to B
      R: mxm rotation matrix
      t: mx1 translation vector
    '''

    # assert A.shape == B.shape

    # get number of dimensions
    m = A.shape[1]

    # translate points to their centroids
    centroid_A = np.mean(A, axis=0)
    centroid_B = np.mean(B, axis=0)
    AA = A - centroid_A
    BB = B - centroid_B

    # rotation matrix
    H = np.dot(AA.T, BB)
    U, S, Vt = np.linalg.svd(H)
    R = np.dot(Vt.T, U.T)

    # special reflection case
    if np.linalg.det(R) < 0:
       Vt[m-1,:] *= -1
       R = np.dot(Vt.T, U.T)

    # translation
    t = centroid_B.T - np.dot(R,centroid_A.T)

    # homogeneous transformation
    T = np.identity(m+1)
    T[:m, :m] = R
    T[:m, m] = t

    return T, R, t

def ICP_align(adata_concat, adata_target, adata_ref, slice_target, slice_ref, landmark_domain, plot_align=False):
    ### find MNN pairs in the landmark domain with knn=1    
    adata_slice1 = adata_target[adata_target.obs['louvain'].isin(landmark_domain)]
    adata_slice2 = adata_ref[adata_ref.obs['louvain'].isin(landmark_domain)]
    
    
    batch_pair = adata_concat[adata_concat.obs['batch_name'].isin([slice_target, slice_ref]) & adata_concat.obs['louvain'].isin(landmark_domain)]
    mnn_dict = create_dictionary_mnn(batch_pair, use_rep='STAligner', batch_name='batch_name', k=1, iter_comb=None, verbose=0)
    adata_1 = batch_pair[batch_pair.obs['batch_name']==slice_target]
    adata_2 = batch_pair[batch_pair.obs['batch_name']==slice_ref]
    
    anchor_list = []
    positive_list = []
    for batch_pair_name in mnn_dict.keys(): 
        for anchor in mnn_dict[batch_pair_name].keys():
            positive_spot = mnn_dict[batch_pair_name][anchor][0]
            ### anchor should only in the ref slice, pos only in the target slice
            if anchor in adata_1.obs_names and positive_spot in adata_2.obs_names:                 
                anchor_list.append(anchor)
                positive_list.append(positive_spot)
                             
    batch_as_dict = dict(zip(list(adata_concat.obs_names), range(0, adata_concat.shape[0])))
    anchor_ind = list(map(lambda _: batch_as_dict[_], anchor_list))
    positive_ind = list(map(lambda _: batch_as_dict[_], positive_list))
    anchor_arr = adata_concat.obsm['STAligner'][anchor_ind, ]
    positive_arr = adata_concat.obsm['STAligner'][positive_ind, ]
    dist_list = [np.sqrt(np.sum(np.square(anchor_arr[ii, :] - positive_arr[ii, :]))) for ii in range(anchor_arr.shape[0])]
    
    
    key_points_src =  np.array(anchor_list)[dist_list < np.percentile(dist_list, 50)] ## remove remote outliers
    key_points_dst =  np.array(positive_list)[dist_list < np.percentile(dist_list, 50)]
    #print(len(anchor_list), len(key_points_src))
    
    coor_src = adata_slice1.obsm["spatial"] ## to_be_aligned
    coor_dst = adata_slice2.obsm["spatial"] ## reference_points

    ## index number
    MNN_ind_src = [list(adata_1.obs_names).index(key_points_src[ii]) for ii in range(len(key_points_src))]
    MNN_ind_dst = [list(adata_2.obs_names).index(key_points_dst[ii]) for ii in range(len(key_points_dst))]
    
    
    ####### ICP alignment
    init_pose = None
    max_iterations = 100
    tolerance = 0.001

    coor_used = coor_src ## Batch_list[1][Batch_list[1].obs['annotation']==2].obsm["spatial"]
    coor_all = adata_target.obsm["spatial"].copy()
    coor_used = np.concatenate([coor_used, np.expand_dims(np.ones(coor_used.shape[0]), axis=1)], axis=1).T    
    coor_all = np.concatenate([coor_all, np.expand_dims(np.ones(coor_all.shape[0]), axis=1)], axis=1).T    
    A = coor_src  ## to_be_aligned
    B = coor_dst  ## reference_points

    m = A.shape[1] # get number of dimensions

    # make points homogeneous, copy them to maintain the originals
    src = np.ones((m + 1, A.shape[0]))
    dst = np.ones((m + 1, B.shape[0]))
    src[:m, :] = np.copy(A.T)
    dst[:m, :] = np.copy(B.T)

    # apply the initial pose estimation
    if init_pose is not None:
        src = np.dot(init_pose, src)
    prev_error = 0

    for ii in range(max_iterations + 1):
        p1 = src[:m, MNN_ind_src].T
        p2 = dst[:m, MNN_ind_dst].T
        T, _, _ = best_fit_transform(src[:m, MNN_ind_src].T,
                                              dst[:m, MNN_ind_dst].T) ## compute the transformation matrix based on MNNs
        import math
        distances = np.mean([math.sqrt(((p1[kk, 0] - p2[kk, 0]) ** 2) + ((p1[kk, 1] - p2[kk, 1]) ** 2))
                             for kk in range(len(p1))])

        # update the current source
        src = np.dot(T, src)
        coor_used = np.dot(T, coor_used)
        coor_all = np.dot(T, coor_all)
        
        # check error
        mean_error = np.mean(distances)
        # print(mean_error)
        if np.abs(prev_error - mean_error) < tolerance:
            break
        prev_error = mean_error
   
    aligned_points = coor_used.T    # MNNs in the landmark_domain
    aligned_points_all = coor_all.T # all points in the slice

    if plot_align:
        import matplotlib.pyplot as plt
        plt.rcParams["figure.figsize"] = (3, 3)
        fig, ax = plt.subplots(1, 2, figsize=(8, 3), gridspec_kw={'wspace': 0.5, 'hspace': 0.1})
        ax[0].scatter(adata_slice2.obsm["spatial"][:, 0], adata_slice2.obsm["spatial"][:, 1],
                    c="blue", cmap=plt.cm.binary_r, s=1)
        ax[0].set_title('Reference '+slice_ref, size=14)
        ax[1].scatter(aligned_points[:, 0], aligned_points[:, 1],
                    c="blue", cmap=plt.cm.binary_r, s=1)
        ax[1].set_title('Target '+slice_target, size=14)

        plt.axis("equal")
        # plt.axis("off")
        plt.show()
        
    #adata_target.obsm["spatial"] = aligned_points_all[:,:2] 
    return aligned_points_all[:,:2]
    
    
    
# https://scikit-learn.org/dev/modules/generated/sklearn.neighbors.NearestNeighbors.html#sklearn.neighbors.NearestNeighbors
def nearest_neighbor(src, dst):
    '''
    Find the nearest (Euclidean) neighbor in dst for each point in src
    Input:
        src: Nxm array of points
        dst: Nxm array of points
    Output:
        distances: Euclidean distances of the nearest neighbor
        indices: dst indices of the nearest neighbor
    '''

    # assert src.shape == dst.shape
    neigh = sklearn.neighbors.NearestNeighbors(n_neighbors=1)
    neigh.fit(dst)
    distances, indices = neigh.kneighbors(src, return_distance=True)
    return distances.ravel(), indices.ravel()





"""
Utility functions for plotting in Palantir
"""

from typing import Optional, Union, Dict, List, Tuple, Any, Callable
import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize
from matplotlib.axes import Axes
from mpl_toolkits.axes_grid1 import make_axes_locatable
import contextlib
import logging


@contextlib.contextmanager
def no_mellon_log_messages():
    # Import mellon locally to avoid JAX fork warnings in other parts of the code
    import mellon
    current_level = mellon.logger.level
    mellon.logger.setLevel(logging.ERROR)
    try:
        yield
    finally:
        mellon.logger.setLevel(current_level)

def _scatter_with_colorbar(
    ax: Axes,
    x: np.ndarray,
    y: np.ndarray,
    c: np.ndarray,
    colorbar_label: Optional[str] = None,
    s: float = 5,
    cmap: Union[str, matplotlib.colors.Colormap] = "viridis",
    norm: Optional[Normalize] = None,
    alpha: float = 1.0,
    **kwargs,
) -> Tuple[Axes, matplotlib.colorbar.Colorbar]:
    """Helper function to create scatter plot with colorbar.

    Parameters
    ----------
    ax : Axes
        Matplotlib axes object to plot on.
    x : np.ndarray
        X-coordinates for scatter plot.
    y : np.ndarray
        Y-coordinates for scatter plot.
    c : np.ndarray
        Values for color mapping.
    colorbar_label : str, optional
        Label for the colorbar. Default is None.
    s : float, optional
        Size of scatter points. Default is 5.
    cmap : str or matplotlib.colors.Colormap, optional
        Colormap for the scatter plot. Default is 'viridis'.
    norm : Normalize, optional
        Normalization for colormap. Default is None.
    alpha : float, optional
        Transparency of scatter points. Default is 1.0.
    **kwargs : dict
        Additional keyword arguments to pass to plt.scatter.

    Returns
    -------
    Tuple[Axes, matplotlib.colorbar.Colorbar]
        The axes object and the colorbar object.
    """
    sc = ax.scatter(x, y, c=c, s=s, cmap=cmap, norm=norm, alpha=alpha, **kwargs)
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = plt.colorbar(sc, cax=cax, orientation="vertical")
    if colorbar_label:
        cbar.set_label(colorbar_label)
    return ax, cbar


def _highlight_cells(
    ax: Axes,
    x: np.ndarray,
    y: np.ndarray,
    mask: np.ndarray,
    deselected_color: str = "lightgray",
    selected_color: str = "crimson",
    s_selected: float = 10,
    s_deselected: float = 3,
    alpha_deselected: float = 0.5,
    alpha_selected: float = 1.0,
    **kwargs,
) -> Axes:
    """Helper function to highlight cells in scatter plot based on mask.

    Parameters
    ----------
    ax : Axes
        Matplotlib axes object to plot on.
    x : np.ndarray
        X-coordinates for scatter plot.
    y : np.ndarray
        Y-coordinates for scatter plot.
    mask : np.ndarray
        Boolean mask for selecting cells to highlight.
    deselected_color : str, optional
        Color for non-highlighted cells. Default is "lightgray".
    selected_color : str, optional
        Color for highlighted cells. Default is "crimson".
    s_selected : float, optional
        Size of highlighted scatter points. Default is 10.
    s_deselected : float, optional
        Size of non-highlighted scatter points. Default is 3.
    alpha_deselected : float, optional
        Transparency of non-highlighted cells. Default is 0.5.
    alpha_selected : float, optional
        Transparency of highlighted cells. Default is 1.0.
    **kwargs : dict
        Additional keyword arguments to pass to plt.scatter.

    Returns
    -------
    Axes
        The modified axes object.
    """
    ax.scatter(
        x[~mask],
        y[~mask],
        c=deselected_color,
        s=s_deselected,
        alpha=alpha_deselected,
        label="Other Cells",
        **kwargs,
    )
    ax.scatter(
        x[mask],
        y[mask],
        c=selected_color,
        s=s_selected,
        alpha=alpha_selected,
        label="Selected Cells",
        **kwargs,
    )
    return ax


def _add_legend(
    ax: Axes,
    handles: Optional[List] = None,
    labels: Optional[List[str]] = None,
    loc: str = "best",
    title: Optional[str] = None,
    **kwargs,
) -> matplotlib.legend.Legend:
    """Helper function to add legend to plot.

    Parameters
    ----------
    ax : Axes
        Matplotlib axes object to add legend to.
    handles : List, optional
        List of artists (lines, patches) to be added to the legend. Default is None.
    labels : List[str], optional
        List of labels for the legend. Default is None.
    loc : str, optional
        Location of the legend. Default is "best".
    title : str, optional
        Title for the legend. Default is None.
    **kwargs : dict
        Additional keyword arguments to pass to ax.legend().

    Returns
    -------
    matplotlib.legend.Legend
        The legend object.
    """
    if handles is not None and labels is not None:
        legend = ax.legend(handles, labels, loc=loc, title=title, **kwargs)
    else:
        legend = ax.legend(loc=loc, title=title, **kwargs)
    return legend


def _setup_axes(
    figsize: Tuple[float, float] = (6, 6),
    ax: Optional[Axes] = None,
    fig: Optional[plt.Figure] = None,
    **kwargs,
) -> Tuple[plt.Figure, Axes]:
    """Helper function to set up figure and axes for plotting.

    Parameters
    ----------
    figsize : Tuple[float, float], optional
        Size of the figure (width, height) in inches. Default is (6, 6).
    ax : Axes, optional
        Existing axes to plot on. Default is None.
    fig : Figure, optional
        Existing figure to plot on. Default is None.
    **kwargs : dict
        Additional keyword arguments to pass to plt.subplots().

    Returns
    -------
    Tuple[plt.Figure, Axes]
        The figure and axes objects.
    """
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize, **kwargs)
    elif fig is None:
        fig = ax.figure
    return fig, ax


def _get_palantir_fates_colors(
    ad,
    fate_names: List[str],
    palantir_fates_colors: Optional[Union[List[str], Dict[str, str]]] = None
) -> Dict[str, str]:
    """
    Generate or update the mapping from branch names to colors.
    
    This utility checks if ad.uns already contains predefined colors.
    Then, if the `palantir_fates_colors` parameter is provided, its values are merged
    (with user-specified colors taking precedence). For any missing branch the function 
    generates a new color ensuring that no color is used twice.
    
    Parameters
    ----------
    ad : AnnData
        The annotated data object from which .uns will be checked.
    fate_names : list of str
        List of branch (fate) names.
    palantir_fates_colors : dict or list or None, optional
        If a dict, keys should be branch names with a color for each.
        If a list, its order is assumed to correspond to fate_names.
        If None, only the predefined colors (if any) and generated defaults are used.

    Returns
    -------
    dict
        Mapping from branch names to colors.
    """
    # Get any predefined colors stored in ad.uns.
    predefined = {}
    if "palantir_fates_colors" in ad.uns:
        predefined = ad.uns["palantir_fates_colors"]
    
    # Process user-provided colors from argument.
    provided = {}
    if palantir_fates_colors is not None:
        if isinstance(palantir_fates_colors, dict):
            provided = palantir_fates_colors
        elif isinstance(palantir_fates_colors, list):
            if len(palantir_fates_colors) < len(fate_names):
                raise ValueError("Provided color list length is less than the number of branch names.")
            provided = {name: clr for name, clr in zip(fate_names, palantir_fates_colors)}
        else:
            raise TypeError("palantir_fates_colors must be a dict, list, or None.")
    
    # Merge: user-provided takes precedence, then predefined.
    mapping = {}
    for branch in fate_names:
        if branch in provided:
            mapping[branch] = provided[branch]
        elif branch in predefined:
            mapping[branch] = predefined[branch]
    
    # Collect already used colors to exclude duplicates.
    used_colors = set(mapping.values())
    
    # Generate colors for missing branches.
    missing = [branch for branch in fate_names if branch not in mapping]
    if missing:
        # Get the default color cycle.
        default_cycle = plt.rcParams['axes.prop_cycle'].by_key().get(
            'color', ['C0', 'C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7', 'C8', 'C9']
        )
        # Create a generator that skips colors already used.
        def color_generator(exclude):
            for clr in default_cycle:
                if clr not in exclude:
                    yield clr
            hex_digits = np.array(list("0123456789ABCDEF"))
            # If default cycle is exhausted, generate random colors.
            while True:
                new_color = "#" + "".join(np.random.choice(hex_digits, size=6))
                if new_color not in exclude:
                    yield new_color

        gen = color_generator(used_colors)
        for branch in missing:
            new_color = next(gen)
            mapping[branch] = new_color
            used_colors.add(new_color)
    
    return mapping


def _plot_arrows(x, y, n=5, ax=None, arrowprops=dict(), arrow_zorder=2, head_offset=0.0, **kwargs):
    """
    Helper function to plot arrows on a trajectory line.
    
    The new 'head_offset' parameter (as a fraction of the segment length)
    moves the arrow head slightly forward.
    
    Parameters
    ----------
    x, y : array-like
        Coordinates of the trajectory points.
    n : int, optional
        Number of arrows to plot. Defaults to 5.
    ax : matplotlib.axes.Axes, optional
        Axes to plot on.
    arrowprops : dict, optional
        Properties for the arrow style.
    arrow_zorder : int, optional
        zorder level for both the line and arrow annotations.
    head_offset : float, optional
        Fraction of the segment length to move the arrow head forward.
    **kwargs :
        Extra keyword arguments passed to the plot function.
    
    Returns
    -------
    matplotlib.axes.Axes
        The axis with the arrows plotted.
    """
    if ax is None:
        fig, ax = plt.subplots()
    
    default_kwargs = {"color": "black", "zorder": arrow_zorder}
    default_kwargs.update(kwargs)
    
    # Plot the trajectory line.
    ax.plot(x, y, **default_kwargs)
    
    if n <= 0:
        return ax
    
    default_arrowprops = dict(arrowstyle="->", lw=1, mutation_scale=20)
    default_arrowprops["color"] = default_kwargs.get("color", "black")
    default_arrowprops.update(arrowprops)
    
    total_points = len(x)
    section_length = total_points // n
    
    for i in range(n):
        idx = total_points - i * section_length
        if idx < 2:
            break
        # Compute the vector from the previous point to the arrow head.
        dx = x[idx - 1] - x[idx - 2]
        dy = y[idx - 1] - y[idx - 2]
        norm = (dx**2 + dy**2) ** 0.5
        # Compute the forward offset.
        if norm != 0:
            offset_dx = head_offset * dx / norm
            offset_dy = head_offset * dy / norm
        else:
            offset_dx = offset_dy = 0
        # Adjust the arrow head coordinates.
        target = (x[idx - 1] + offset_dx, y[idx - 1] + offset_dy)
        
        ax.annotate(
            "",
            xy=target,
            xytext=(x[idx - 2], y[idx - 2]),
            arrowprops=default_arrowprops,
            zorder=arrow_zorder,
        )
    return ax


r"""
utils (Pyomic.utils)

"""

'''
from ._data import (read,
                    read_csv,
                    read_10x_mtx,
                    read_h5ad,
                    read_10x_h5,
                    data_downloader,
                    download_CaDRReS_model,
                    download_GDSC_data,
                    download_pathway_database,
                    download_geneid_annotation_pair,
                    download_tosica_gmt,
                    geneset_prepare,
                    get_gene_annotation,
                    correlation_pseudotime,
                    np_mean,
                    np_std,
                    anndata_sparse,
                    store_layers,
                    retrieve_layers,
                    easter_egg
                    )'''
from ._data import *
from ._plot import *
#from ._genomics import *
from ._mde import *
from ._syn import *
from ._scatterplot import *
from ._knn import *
from ._heatmap import *
from ._roe import roe
from ._paga import cal_paga,plot_paga
from ._cluster import cluster,LDA_topic,filtered,refine_label
from ._venn import venny4py
from ._lsi import *
from ._neighboors import neighbors



import pandas as pd
import numpy as np
import anndata
import scanpy as sc
import torch
from typing import Union

from ..bulk import data_drop_duplicates_index,deseq2_normalize
from ..bulk2single import Bulk2Single
from ..single import scnocd
from ._vae import train_vae, generate_vae, load_vae

class BulkTrajBlend(object):
    """
    BulkTrajBlend: A class for bulk and single cell data integration and trajectory inference using beta-VAE and GNN.


    """

    def __init__(self,bulk_seq:pd.DataFrame,single_seq:anndata.AnnData,
                 celltype_key:str,bulk_group=None,max_single_cells:int=5000,
                 top_marker_num:int=500,ratio_num:int=1,gpu:Union[int,str]=0) -> None:
        """
        Initialize the BulkTrajBlend class

        Arguments:
            bulk_seq: The bulk data. The index is gene name and the columns is cell name.
            single_seq: The single cell data. The index is cell name and the columns is gene name.
            celltype_key: The key of cell type in the single cell data.
            top_marker_num: The number of top marker genes for each cell type.
            ratio_num: The number of cells to be selected for each cell type.
            gpu: The gpu id or 'cpu' or 'mps'.
            max_single_cells: The maximum number of single cells to be used. Default is 5000.

        """

        self.bulk_seq = bulk_seq.copy()
        self.single_seq = single_seq.copy()
        self.celltype_key=celltype_key
        self.top_marker_num=top_marker_num
        self.ratio_num=ratio_num
        self.gpu=gpu
        self.group=bulk_group
        self.max_single_cells=max_single_cells
        if gpu=='mps' and torch.backends.mps.is_available():
            print('Note that mps may loss will be nan, used it when torch is supported')
            self.used_device = torch.device("mps")
        else:
            self.used_device = torch.device(f"cuda:{gpu}") if gpu >= 0 and torch.cuda.is_available() else torch.device('cpu')
        self.history=[]
        data_dg_v=self.bulk_seq.mean(axis=1)
        data_dg=pd.DataFrame(index=data_dg_v.index)
        data_dg['group']=data_dg_v
        self.bulk_seq_group=data_dg
        pass

    def bulk_preprocess_lazy(self,)->None:
        """
        Preprocess the bulk data

        Arguments:
            group: The group of the bulk data. Default is None. It need to set to calculate the mean of each group.
        """

        print("......drop duplicates index in bulk data")
        self.bulk_seq=data_drop_duplicates_index(self.bulk_seq)
        print("......deseq2 normalize the bulk data")
        self.bulk_seq=deseq2_normalize(self.bulk_seq)
        print("......log10 the bulk data")
        self.bulk_seq=np.log10(self.bulk_seq+1)
        print("......calculate the mean of each group")
        if self.group is None:
            return None
        else:
            data_dg_v=self.bulk_seq[self.group].mean(axis=1)
            data_dg=pd.DataFrame(index=data_dg_v.index)
            data_dg['group']=data_dg_v
            self.bulk_seq_group=data_dg
        return None
    
    def single_preprocess_lazy(self,target_sum:int=1e4)->None:
        """
        Preprocess the single data

        Arguments:
            target_sum: The target sum of the normalize. Default is 1e4.

        """

        print("......normalize the single data")
        self.single_seq.obs_names_make_unique()
        self.single_seq.var_names_make_unique()
        sc.pp.normalize_total(self.single_seq, target_sum=target_sum)
        print("......log1p the single data")
        sc.pp.log1p(self.single_seq)
        return None
    
    def vae_configure(self,cell_target_num=None,**kwargs):
        """
        Configure the vae model

        Arguments:
            cell_target_num: The number of cell types to be generated. Default is 100.

        
        """
        self.vae_model=Bulk2Single(bulk_data=self.bulk_seq,single_data=self.single_seq,
                                   celltype_key=self.celltype_key,bulk_group=self.group,
                                      max_single_cells=self.max_single_cells,
                 top_marker_num=self.top_marker_num,ratio_num=self.ratio_num,gpu=self.gpu)
        if cell_target_num!=None:
            self.vae_model.cell_target_num=dict(zip(list(set(self.single_seq.obs[self.celltype_key])),
                                                [cell_target_num]*len(list(set(self.single_seq.obs[self.celltype_key])))))
        else:
            self.cellfract=self.vae_model.predicted_fraction(**kwargs)
        
        self.sc_ref=self.vae_model.sc_ref.copy()
        self.bulk_ref=self.vae_model.bulk_data.T.copy()

        self.vae_model.bulk_preprocess_lazy()
        self.vae_model.single_preprocess_lazy()
        self.vae_model.prepare_input()

        

    def vae_train(self,
                  vae_save_dir:str='save_model',
            vae_save_name:str='vae',
            generate_save_dir:str='output',
            generate_save_name:str='output',
            batch_size:int=512,
            learning_rate:int=1e-4,
            hidden_size:int=256,
            epoch_num:int=5000,
            patience:int=50,save:bool=True):
        r"""
        Train the VAE model of BulkTrajBlend.

        Arguments:
            vae_save_dir: The directory to save the trained VAE model. Default is 'save_model'.
            vae_save_name: The name to save the trained VAE model. Default is 'vae'.
            generate_save_dir: The directory to save the generated single-cell data. Default is 'output'.
            generate_save_name: The name to save the generated single-cell data. Default is 'output'.
            batch_size: The batch size for training the VAE model. Default is 512.
            learning_rate: The learning rate for training the VAE model. Default is 1e-4.
            hidden_size: The hidden size for the encoder and decoder networks. Default is 256.
            epoch_num: The epoch number for training the VAE model. Default is 5000.
            patience: The patience for training the VAE model. Default is 50.
            save: Whether to save the trained VAE model. Default is True.

        """
        
        self.vae_net=self.vae_model.train(
                batch_size=batch_size,
                learning_rate=learning_rate,
                hidden_size=hidden_size,
                epoch_num=epoch_num,
                vae_save_dir=vae_save_dir,
                vae_save_name=vae_save_name,
                generate_save_dir=generate_save_dir,
                generate_save_name=generate_save_name,
                patience=patience,save=save)
        
        
    def vae_load(self,vae_load_dir:str,hidden_size:int=256):
        r"""
        load the trained VAE model of BulkTrajBlend.

        Arguments:
            vae_load_dir: The directory to load the trained VAE model.
            hidden_size: The hidden size for the encoder and decoder networks. Default is 256.
        """

        print(f'loading model from {vae_load_dir}')
        vae_net = self.vae_model.load(vae_load_dir,hidden_size=hidden_size)
        self.vae_net=vae_net

    def vae_generate(self,highly_variable_genes:bool=True,max_value:float=10,
                     n_comps:int=100,svd_solver:str='auto',leiden_size:int=50)->anndata.AnnData:
        """
        Generate the single-cell data from the trained VAE model.

        Arguments:
            highly_variable_genes: Whether to use highly variable genes. Default is True.
            max_value: The maximum value for the scaled data. Default is 10.
            n_comps: The number of principal components. Default is 100.
            svd_solver: The solver for the PCA. Default is 'auto'.
            leiden_size: The minimum size of the leiden clusters. Default is 50.

        Returns:
            generate_adata: The generated single-cell data.

        """

        generate_adata=self.vae_model.generate()
        self.generate_adata_raw=generate_adata.copy()
        generate_adata.raw = generate_adata
        if highly_variable_genes:
            sc.pp.highly_variable_genes(generate_adata, min_mean=0.0125, max_mean=3, min_disp=0.5)
            generate_adata = generate_adata[:, generate_adata.var.highly_variable]
        sc.pp.scale(generate_adata, max_value=max_value)
        sc.tl.pca(generate_adata, n_comps=n_comps, svd_solver=svd_solver)
        sc.pp.neighbors(generate_adata, use_rep="X_pca")
        sc.tl.leiden(generate_adata)
        filter_leiden=list(generate_adata.obs['leiden'].value_counts()[generate_adata.obs['leiden'].value_counts()<leiden_size].index)
        generate_adata.uns['noisy_leiden']=filter_leiden
        print("The filter leiden is ",filter_leiden)
        generate_adata=generate_adata[~generate_adata.obs['leiden'].isin(filter_leiden)]
        self.generate_adata=generate_adata.copy()
        return generate_adata.raw.to_adata()
    
    def gnn_configure(self,use_rep='X',neighbor_rep='X_pca',
                      gpu=0,hidden_size:int=128,
                     weight_decay:int=1e-2,
                     dropout:float=0.5,
                     batch_norm:bool=True,
                     lr:int=1e-3,
                     max_epochs:int=500,
                     display_step:int=25,
                     balance_loss:bool=True,
                     stochastic_loss:bool=True,
                     batch_size:int=2000,num_workers:int=5,):
        """
        Configure the GNN model of BulkTrajBlend.

        Arguments:
            gpu: The GPU ID for training the GNN model. Default is 0.
            hidden_size: The hidden size for the GNN model. Default is 128.
            weight_decay: The weight decay for the GNN model. Default is 1e-2.
            dropout: The dropout for the GNN model. Default is 0.5.
            batch_norm: Whether to use batch normalization for the GNN model. Default is True.
            lr: The learning rate for the GNN model. Default is 1e-3.
            max_epochs: The maximum epoch number for training the GNN model. Default is 500.
            display_step: The display step for training the GNN model. Default is 25.
            balance_loss: Whether to use the balance loss for training the GNN model. Default is True.
            stochastic_loss: Whether to use the stochastic loss for training the GNN model. Default is True.
            batch_size: The batch size for training the GNN model. Default is 2000.
            num_workers: The number of workers for training the GNN model. Default is 5.


        """
        nocd_obj=scnocd(self.generate_adata,use_rep=use_rep,
                        neighbor_rep=neighbor_rep,gpu=gpu)
        #nocd_obj.device = torch.device(f"cuda:{gpu}") if gpu >= 0 and torch.cuda.is_available() else torch.device('cpu')
        nocd_obj.matrix_transform(clustertype=self.celltype_key)
        nocd_obj.matrix_normalize()
        nocd_obj.GNN_configure(hidden_size=hidden_size,weight_decay=weight_decay,
                             dropout=dropout,batch_norm=batch_norm,lr=lr,
                             max_epochs=max_epochs,display_step=display_step,
                             balance_loss=balance_loss,stochastic_loss=stochastic_loss,
                             batch_size=batch_size)
        nocd_obj.GNN_preprocess(num_workers=num_workers)

        self.nocd_obj=nocd_obj
    
    def gnn_train(self,thresh:float=0.5,gnn_save_dir:str='save_model',
            gnn_save_name:str='gnn'):
        """
        Train the GNN model of BulkTrajBlend.

        Arguments:
            thresh: The threshold for the GNN model. Default is 0.5.
            gnn_save_dir: The directory for saving the GNN model. Default is 'save_model'.
            gnn_save_name: The name for saving the GNN model. Default is 'gnn'.
        
        """
        self.nocd_obj.GNN_model()
        self.nocd_obj.GNN_result(thresh=thresh)
        self.nocd_obj.cal_nocd()
        self.nocd_obj.save(gnn_save_dir=gnn_save_dir,gnn_save_name=gnn_save_name)

    def gnn_load(self,gnn_load_dir:str,thresh:float=0.5,):
        """
        Load the GNN model of BulkTrajBlend.

        Arguments:
            gnn_load_dir: The directory for loading the GNN model.
            thresh: The threshold for the GNN model. Default is 0.5.
        
        """
        self.nocd_obj.load(gnn_load_dir)
        self.nocd_obj.GNN_result(thresh=thresh)
        self.nocd_obj.cal_nocd()
    
    def gnn_generate(self)->pd.DataFrame:
        """
        Generate the overlap cell community.

        Returns:
            res_pd: The overlap cell community.
        
        """
        '''
        pair_dict_r={}
        for i in range(self.nocd_obj.K):
            j=0
            while 1:
                if self.nocd_obj.adata[self.nocd_obj.adata.obs['nocd_n']==str(i)].shape[0]==0:
                    break
                if j>=len(self.nocd_obj.adata[self.nocd_obj.adata.obs['nocd_n']==str(i)].obs.value_counts(self.celltype_key).index):
                    pair_dict_r[str(i)]=self.nocd_obj.adata[self.nocd_obj.adata.obs['nocd_n']==str(i)].obs.value_counts(self.celltype_key).index[j-1]+'_'+str(j)
                    break
                if self.nocd_obj.adata[self.nocd_obj.adata.obs['nocd_n']==str(i)].obs.value_counts(self.celltype_key).index[j] not in list(pair_dict_r.values()):
                    pair_dict_r[str(i)]=self.nocd_obj.adata[self.nocd_obj.adata.obs['nocd_n']==str(i)].obs.value_counts(self.celltype_key).index[j]
                    break
                else:
                    j+=1
        pair_dict_r
        '''
        unique_adata=self.nocd_obj.adata[~self.nocd_obj.adata.obs['nocd_n'].str.contains('-')]
        pair_dict_r={}
        repeat_celltype=dict(zip(list(set(unique_adata.obs[self.celltype_key])),np.zeros(len(list(set(unique_adata.obs[self.celltype_key]))))))
        for nocd_class in list(set(unique_adata.obs['nocd_n'])):
            now_celltype=unique_adata[unique_adata.obs['nocd_n']==nocd_class].obs.value_counts(self.celltype_key).index[0]
            if (now_celltype in pair_dict_r.values()):
                #print(now_celltype)
                pair_dict_r[str(nocd_class)]=now_celltype+'_'+str(int(repeat_celltype[now_celltype]))
                repeat_celltype[now_celltype]+=1
            else:
                pair_dict_r[str(nocd_class)]=now_celltype
                repeat_celltype[now_celltype]+=1

        def li_range(li,max_len):
            r=[0]*max_len   
            for i in li:
                r[int(i)]=1
            return r
        
        res_li=[li_range(i.split('-'),self.nocd_obj.K) for i in self.nocd_obj.adata.obs['nocd_n']]
        res_pd=pd.DataFrame(res_li,index=self.nocd_obj.adata.obs.index,columns=['nocd_'+i for i in [pair_dict_r[str(j)] for j in range(self.nocd_obj.K)]])
        print("The nocd result is ",res_pd.sum(axis=0))
        print("The nocd result has been added to adata.obs['nocd_']")
        self.nocd_obj.adata.obs=pd.concat([self.nocd_obj.adata.obs,res_pd],axis=1)
        return res_pd 
    
    def interpolation(self,celltype:str,adata:anndata.AnnData=None,)->anndata.AnnData:
        """
        Interpolate the cell community to raw data.

        Arguments:
            celltype: The cell type for interpolation.
            adata: The raw data for interpolation. If is None, will use the single_seq data. Default is None.

        Returns:
            adata1: The adata after interpolated .
        """
        if adata is None:
            adata=self.single_seq
        test_adata=self.nocd_obj.adata[self.nocd_obj.adata.obs['nocd_{}'.format(celltype)]==1].raw.to_adata()
        if test_adata.shape[0]!=0:
            adata1=anndata.concat([test_adata,
                        adata],merge='same')
        else:
            adata1=adata 
            print("The cell type {} is not in the nocd result".format(celltype))
        return adata1
    






"""
__author__ = "Xiang Zhou"
__email__ = "xzhou@amss.ac.cn"
__citation__ = Zhou, X., Dong, K. & Zhang, S. Integrating spatial transcriptomics data 
across different conditions, technologies and developmental stages. Nat Comput Sci 3, 894–906 (2023)
"""
from ..externel.STAligner.mnn_utils import create_dictionary_mnn
from ..externel.STAligner.STALIGNER import STAligner

import numpy as np
import pandas as pd
from tqdm import tqdm
import scipy.sparse as sp
import sklearn.neighbors
import random

import torch
import torch.backends.cudnn as cudnn
cudnn.deterministic = True
cudnn.benchmark = True

import torch.nn.functional as F
from torch_geometric.data import Data
from torch_geometric.loader import DataLoader

from .._settings import add_reference


def Cal_Spatial_Net(adata, rad_cutoff=None, k_cutoff=None,
                    max_neigh=50, model='Radius', verbose=True):
    """\
    Construct the spatial neighbor networks.

    Parameters
    ----------
    adata
        AnnData object of scanpy package.
    rad_cutoff
        radius cutoff when model='Radius'
    k_cutoff
        The number of nearest neighbors when model='KNN'
    model
        The network construction model. When model=='Radius', 
        the spot is connected to spots whose distance is less than rad_cutoff.
        When model=='KNN', the spot is connected to its first k_cutoff nearest neighbors.

    Returns
    -------
    The spatial networks are saved in adata.uns['Spatial_Net']

    Example
    -------
    >>> ov.space.Cal_Spatial_Net(adata, rad_cutoff=50, model='Radius')
    or
    >>> ov.space.Cal_Spatial_Net(adata, k_cutoff=10, model='KNN')
    When using STAligner, it is necessary to adjust the rad_cutoff parameter according to 
    different data to ensure that each spot has an average of 5-10 adjacent spots connected to it. 
    Such as: "11.3356 neighbors per cell on average."
    """

    assert (model in ['Radius', 'KNN'])
    if verbose:
        print('------Calculating spatial graph...')
    coor = pd.DataFrame(adata.obsm['spatial'])
    coor.index = adata.obs.index
    coor.columns = ['imagerow', 'imagecol']


    nbrs = sklearn.neighbors.NearestNeighbors(
        n_neighbors=max_neigh + 1, algorithm='ball_tree').fit(coor)
    distances, indices = nbrs.kneighbors(coor)

    if model == 'KNN':
        indices = indices[:, 1:k_cutoff + 1]
        distances = distances[:, 1:k_cutoff + 1]
    if model == 'Radius':
        indices = indices[:, 1:]
        distances = distances[:, 1:]

    KNN_list = []
    for it in range(indices.shape[0]):
        KNN_list.append(pd.DataFrame(zip([it] * indices.shape[1], indices[it, :],distances[it, :])))
    KNN_df = pd.concat(KNN_list)
    KNN_df.columns = ['Cell1', 'Cell2', 'Distance']

    Spatial_Net = KNN_df.copy()
    if model == 'Radius':
        Spatial_Net = KNN_df.loc[KNN_df['Distance'] < rad_cutoff,]
    id_cell_trans = dict(zip(range(coor.shape[0]), np.array(coor.index), ))
    Spatial_Net['Cell1'] = Spatial_Net['Cell1'].map(id_cell_trans)
    Spatial_Net['Cell2'] = Spatial_Net['Cell2'].map(id_cell_trans)

    # self_loops = pd.DataFrame(zip(Spatial_Net['Cell1'].unique(), Spatial_Net['Cell1'].unique(),
    #                  [0] * len((Spatial_Net['Cell1'].unique())))) ###add self loops
    # self_loops.columns = ['Cell1', 'Cell2', 'Distance']
    # Spatial_Net = pd.concat([Spatial_Net, self_loops], axis=0)

    if verbose:
        print(f'The graph contains {Spatial_Net.shape[0]} edges, {adata.n_obs} cells.')
        print(f'{(Spatial_Net.shape[0] / adata.n_obs):.4f}neighbors per cell on average.')
    adata.uns['Spatial_Net'] = Spatial_Net

    #########
    #X = pd.DataFrame(adata.X.toarray()[:, ], index=adata.obs.index, columns=adata.var.index) 
    #cells = np.array(X.index)
    cells = np.array(adata.obs_names) # LeiHu update, reducing the demand of memory.
    cells_id_tran = dict(zip(cells, range(cells.shape[0])))
    if 'Spatial_Net' not in adata.uns.keys():
        raise ValueError("Spatial_Net is not existed! Run Cal_Spatial_Net first!")
    Spatial_Net = adata.uns['Spatial_Net']
    G_df = Spatial_Net.copy()
    G_df['Cell1'] = G_df['Cell1'].map(cells_id_tran)
    G_df['Cell2'] = G_df['Cell2'].map(cells_id_tran)
    G = sp.coo_matrix((np.ones(G_df.shape[0]), (G_df['Cell1'], G_df['Cell2'])), shape=(adata.n_obs, adata.n_obs))
    G = G + sp.eye(G.shape[0])  # self-loop
    adata.uns['adj'] = G

class pySTAligner(object):
    def __init__(self,adata,
                 hidden_dims: list = [512, 30],
                 n_epochs: int = 1000,
                 lr: float = 0.001,
                 batch_key: str = 'batch_name',
                 key_added: str = 'STAligner',
                 gradient_clipping: float = 5,
                 weight_decay: float = 0.0001,
                 margin: float = 1,
                 verbose: bool = False,
                 random_seed: int = 666,
                 iter_comb = None,
                 knn_neigh: int = 100,
                 Batch_list = None,
                 device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
             ) -> None:
        r'''
        Parameter
        ----------
        adata: AnnData
            The AnnData object of your input data.
        hidden_dims: list
            The hidden dimensions of the neural network.
        n_epochs: int   
            The number of epochs for training.  
        lr: float   
            The learning rate for training.
        batch_key: str  
            The key of batch information in adata.obs.
        key_added: str  
            The key of the new embedding in adata.obsm.
        gradient_clipping: float    
            The gradient clipping value.
        weight_decay: float
            The weight decay value.
        margin: float
            The margin value for triplet loss.
        verbose: bool
            Whether to print the training process.
        random_seed: int
            The random seed for training.
        iter_comb: list
            The list of pairwise comparison of batches.
        knn_neigh: int
            The number of nearest neighbors for MNN.
        Batch_list: list
            The list of AnnData objects of each batch.
        device: torch.device
            The device for training.

        Example
        ----------
        >>> STAligner_obj = ov.space.pySTAligner(adata_concat, verbose=True, knn_neigh = 100,
                                     n_epochs = 600, iter_comb = iter_comb,
                                     batch_key = 'batch_name',  key_added='STAligner',
                                     Batch_list = Batch_list)
        '''
        self.device = device
        section_ids = np.array(adata.obs[batch_key].unique())

        comm_gene = adata.var_names
        data_list = []
        for adata_tmp in Batch_list:
            adata_tmp = adata_tmp[:, comm_gene]
            edge_index = np.nonzero(adata_tmp.uns['adj'])
            data_list.append(
                Data(edge_index=torch.LongTensor(np.array([edge_index[0], edge_index[1]])),
                              prune_edge_index=torch.LongTensor(np.array([])),
                              x=torch.FloatTensor(adata_tmp.X.todense())))

        loader = DataLoader(data_list, batch_size=1, shuffle=True)

        self.loader=loader
        self.adata = adata
        self.data_list = data_list

        # hyper-parameters
        self.lr=lr
        self.section_ids = section_ids
        self.n_epochs = n_epochs
        self.weight_decay=weight_decay
        self.hidden_dims = hidden_dims
        self.key_added = key_added
        self.gradient_clipping = gradient_clipping
        self.random_seed = random_seed
        self.margin = margin
        self.verbose = verbose
        self.iter_comb = iter_comb
        self.knn_neigh = knn_neigh
        self.Batch_list = Batch_list
        self.batch_key = batch_key
        self.model = STAligner(hidden_dims=[adata.X.shape[1], hidden_dims[0],
                                            hidden_dims[1]]).to(device)
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=lr,
                                          weight_decay=weight_decay)

        if verbose:
            print(self.model)

    def train(self):
        """Train the STAligner model."""  
        seed = self.random_seed
        random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        np.random.seed(seed)

        print('Pretrain with STAGATE...')
        for epoch in tqdm(range(0, 500)):
            for batch in self.loader:
                self.model.train()
                self.optimizer.zero_grad()
                batch = batch.to(self.device)
                z, out = self.model(batch.x, batch.edge_index)

                loss = F.mse_loss(batch.x, out)  # +adv_loss
                loss.backward()
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), 5.)
                self.optimizer.step()


        with torch.no_grad():
            z_list = []
            for batch in self.data_list:
                z, _ = self.model.cpu()(batch.x, batch.edge_index)
                z_list.append(z.cpu().detach().numpy())
        self.adata.obsm['STAGATE'] = np.concatenate(z_list, axis=0)
        self.model = self.model.to(self.device)


        print('Train with STAligner...')
        for epoch in tqdm(range(500, self.n_epochs)):
            if epoch % 100 == 0 or epoch == 500:
                if self.verbose:
                    print('Update spot triplets at epoch ' + str(epoch))
                with torch.no_grad():
                    z_list = []
                    for batch in self.data_list:
                        z, _ = self.model.cpu()(batch.x, batch.edge_index)
                        z_list.append(z.cpu().detach().numpy())

                self.adata.obsm['STAGATE'] = np.concatenate(z_list, axis=0)
                self.model = self.model.to(self.device)

                pair_data_list = []

                for comb in self.iter_comb:
                    #print(comb)
                    i, j = comb[0], comb[1]
                    batch_pair = self.adata[self.adata.obs[self.batch_key].isin([self.section_ids[i],
                                                                                  self.section_ids[j]])]
                    mnn_dict = create_dictionary_mnn(batch_pair, use_rep='STAGATE', batch_name=self.batch_key,
                                                           k=self.knn_neigh,
                                                           iter_comb=None, verbose=0)

                    batchname_list = batch_pair.obs[self.batch_key]
                    cellname_by_batch_dict = dict()
                    for batch_id in range(len(self.section_ids)):
                        cellname_by_batch_dict[self.section_ids[batch_id]] = batch_pair.obs_names[
                            batch_pair.obs[self.batch_key] == self.section_ids[batch_id]].values
                    anchor_list = []
                    positive_list = []
                    negative_list = []
                    for batch_pair_name in mnn_dict.keys():  # pairwise compare for multiple batches
                        for anchor in mnn_dict[batch_pair_name].keys():
                            anchor_list.append(anchor)
                            positive_spot = mnn_dict[batch_pair_name][anchor][0]
                            positive_list.append(positive_spot)
                            section_size = len(cellname_by_batch_dict[batchname_list[anchor]])
                            negative_list.append(
                                cellname_by_batch_dict[batchname_list[anchor]][np.random.randint(section_size)])

                    batch_as_dict = dict(zip(list(batch_pair.obs_names),
                                              range(0, batch_pair.shape[0])))
                    anchor_ind = list(map(lambda _: batch_as_dict[_], anchor_list))
                    positive_ind = list(map(lambda _: batch_as_dict[_], positive_list))
                    negative_ind = list(map(lambda _: batch_as_dict[_], negative_list))

                    edge_list_1 = np.nonzero(self.Batch_list[i].uns['adj'])

                    max_ind = edge_list_1[0].max()
                    edge_list_2 = np.nonzero(self.Batch_list[j].uns['adj'])

                    edge_list_2 = (edge_list_2[0] + max_ind + 1, edge_list_2[1] + max_ind + 1)
                    edge_list = [edge_list_1, edge_list_2]

                    edge_pairs = [np.append(edge_list[0][0], edge_list[1][0]),
                                   np.append(edge_list[0][1], edge_list[1][1])]

                    pair_data_list.append(Data(edge_index=torch.LongTensor(np.array([edge_pairs[0], edge_pairs[1]])),
                                           anchor_ind=torch.LongTensor(np.array(anchor_ind)),
                                           positive_ind=torch.LongTensor(np.array(positive_ind)),
                                           negative_ind=torch.LongTensor(np.array(negative_ind)),
                                           x=batch_pair.X)) 
                    #torch.FloatTensor(batch_pair.X.todense())

                # for temp in pair_data_list:
                #     temp.to(device)
                pair_loader = DataLoader(pair_data_list, batch_size=1, shuffle=True)

            for batch in pair_loader:
                self.model.train()
                self.optimizer.zero_grad()
                batch.x = torch.FloatTensor(batch.x[0].todense())
                batch = batch.to(self.device)
                z, out = self.model(batch.x, batch.edge_index)
                mse_loss = F.mse_loss(batch.x, out)

                anchor_arr = z[batch.anchor_ind,]
                positive_arr = z[batch.positive_ind,]
                negative_arr = z[batch.negative_ind,]

                triplet_loss = torch.nn.TripletMarginLoss(margin=self.margin, p=2, reduction='sum')
                tri_output = triplet_loss(anchor_arr, positive_arr, negative_arr)

                loss = mse_loss + tri_output
                loss.backward()
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.gradient_clipping)
                self.optimizer.step()

        add_reference(self.adata,'STAligner','spatial integration with STAligner')

    def predicted(self):
        """store the embedding of STAligner model."""  
        self.model.eval()
        with torch.no_grad():
            z_list = []
            for batch in self.data_list:
                z, _ = self.model.cpu()(batch.x, batch.edge_index)
                z_list.append(z.cpu().detach().numpy())

        self.adata.obsm[self.key_added] = np.concatenate(z_list, axis=0)
        add_reference(self.adata,'STAligner','spatial integration with STAligner')
        return self.adata
# End-of-file (EOF)



#!/usr/bin/env python3
"""
OmicVerse Analytics Sender
独立的analytics数据发送工具，不依赖HTML报告生成
模拟lzy_tt.html中成功的发送格式
"""

import requests
import platform
import hashlib
import os
from datetime import datetime
from urllib.parse import urlencode
import uuid


def get_geolocation():
    """获取完整的地理位置信息"""
    try:
        response = requests.get('http://ip-api.com/json/', timeout=5)
        if response.status_code == 200:
            geo_data = response.json()
            # 返回完整的地理位置数据
            return geo_data
    except:
        pass
    return {
        'status': 'fail',
        'country': 'Unknown',
        'countryCode': 'XX',
        'region': 'Unknown',
        'regionName': 'Unknown',
        'city': 'Unknown',
        'timezone': 'UTC',
        'isp': 'Unknown',
        'query': 'Unknown'
    }


def generate_user_hash():
    """生成匿名用户hash"""
    machine_info = f"{platform.node()}-{platform.system()}-{platform.machine()}"
    return hashlib.md5(machine_info.encode()).hexdigest()[:8]


def send_analytics(analytics_id, event_type='report_view', **kwargs):
    """
    发送analytics数据到服务器 - 标准版
    
    Parameters:
    -----------
    analytics_id : str
        项目或报告的唯一标识符
    event_type : str
        事件类型，默认为'report_view'
    **kwargs : dict
        额外的参数
        
    Returns:
    --------
    bool : 发送是否成功
    
    Note:
    -----
    包含以下标准字段：
    - id, user, ts (核心字段)
    - platform, ua, lang, tz, country (环境信息)
    - ref (引用页面，Python环境为空)
    
    与send_analytics_full的区别：
    - 标准版：9个基础字段
    - 完整版：17个字段，包含详细地理信息(城市、ISP、坐标等)
    """
    
    # 基础数据收集
    user_hash = generate_user_hash()
    timestamp = datetime.now().isoformat()
    platform_info = platform.system() + " " + platform.release()
    
    # 获取地理位置
    print("🌍 正在获取地理位置...")
    geo_data = get_geolocation()
    print(f"🌍 检测到国家: {geo_data.get('country', 'Unknown')}")
    if geo_data.get('city', 'Unknown') != 'Unknown':
        print(f"🏙️ 城市: {geo_data.get('city')}, {geo_data.get('regionName')}")
    if geo_data.get('timezone', 'UTC') != 'UTC':
        print(f"🕐 时区: {geo_data.get('timezone')}")
    
    # 构建数据包 - 使用服务器期望的参数名称
    data = {
        # 核心字段  
        'id': analytics_id,
        'user': user_hash,
        'ts': timestamp,
        
        # 基础环境信息
        'platform': platform_info,
        'ua': f'Python-OmicVerse/{platform.python_version()}',
        'lang': 'en-US',  # 可以从环境变量获取
        'tz': geo_data.get('timezone', 'UTC'),
        'country': geo_data.get('country', 'Unknown'),
        'ref': '',        # Python环境没有referrer
        
        **kwargs          # 额外参数
    }
    
    # 只在有意义的情况下添加可选字段
    if event_type != 'report_view':
        data['event_type'] = event_type
    
    print(f"📊 发送Analytics数据: {analytics_id}")
    print(f"👤 用户Hash: {user_hash}")
    print(f"🌍 国家: {data['country']}")
    print(f"🕐 时区: {data['tz']}")
    print(f"🌐 语言: {data['lang']}")
    print(f"💻 平台: {platform_info}")
    print(f"⏰ 时间: {timestamp}")
    
    # 发送请求 - 模拟lzy_tt.html中的方式
    analytics_endpoint = os.environ.get('OMICVERSE_ANALYTICS_ENDPOINT', 'http://8.130.139.217/track.gif')
    
    try:
        # 方法1: 使用GET请求（模拟img.src方式）
        params = urlencode(data)
        url = f"{analytics_endpoint}?{params}"
        
        print(f"📡 发送到: {analytics_endpoint}")
        print(f"📋 参数数量: {len(data)} 个字段")
        print(f"🔍 关键参数: lang={data['lang']}, tz={data['tz']}, country={data['country']}")
        
        response = requests.get(url, timeout=10)
        
        if response.status_code == 200:
            print("✅ Analytics数据发送成功!")
            return True
        else:
            print(f"❌ 发送失败: HTTP {response.status_code}")
            return False
            
    except Exception as e:
        print(f"❌ 发送失败: {str(e)}")
        return False


def send_analytics_simple(analytics_id):
    """
    简化版发送函数 - 使用默认参数
    
    Parameters:
    -----------
    analytics_id : str
        项目标识符
        
    Note:
    -----
    实际上调用send_analytics，包含基础环境信息：
    - 项目ID、用户hash、时间戳
    - 平台信息、用户代理、语言、时区、国家
    """
    return send_analytics(analytics_id)


def send_analytics_detailed(analytics_id, project_name=None, user_note=None):
    """
    详细版发送函数，包含额外信息
    
    Parameters:
    -----------
    analytics_id : str
        项目标识符
    project_name : str
        项目名称
    user_note : str
        用户备注
    """
    extra_data = {}
    if project_name:
        extra_data['project_name'] = project_name
    if user_note:
        extra_data['user_note'] = user_note
        
    return send_analytics(analytics_id, **extra_data)


def send_analytics_full(analytics_id, event_type='report_view', **kwargs):
    """
    完整版发送函数，包含所有地理位置信息
    
    Parameters:
    -----------
    analytics_id : str
        项目或报告的唯一标识符
    event_type : str
        事件类型，默认为'report_view'
    **kwargs : dict
        额外的参数
        
    Returns:
    --------
    bool : 发送是否成功
    
    Note:
    -----
    使用与ip-api.com完全一致的字段名称和数据结构
    """
    
    # 基础数据收集
    user_hash = generate_user_hash()
    timestamp = datetime.now().isoformat()
    platform_info = platform.system() + " " + platform.release()
    
    # 获取完整地理位置
    print("🌍 正在获取完整地理位置信息...")
    geo_data = get_geolocation()
    
    # 显示详细信息
    if geo_data.get('status') == 'success':
        print(f"✅ 地理位置获取成功:")
        print(f"   🌍 国家: {geo_data.get('country')} ({geo_data.get('countryCode')})")
        print(f"   🏙️ 城市: {geo_data.get('city')}, {geo_data.get('regionName')}")
        print(f"   📍 地区: {geo_data.get('region')} ({geo_data.get('regionName')})")
        print(f"   📮 邮编: {geo_data.get('zip', 'N/A')}")
        print(f"   🕐 时区: {geo_data.get('timezone')}")
        print(f"   🌐 ISP: {geo_data.get('isp')}")
        print(f"   🏢 组织: {geo_data.get('org')}")
        print(f"   📍 坐标: {geo_data.get('lat')}, {geo_data.get('lon')}")
        print(f"   🔌 AS: {geo_data.get('as')}")
        print(f"   🌐 查询IP: {geo_data.get('query')}")
    else:
        print(f"⚠️ 地理位置获取失败，使用默认值")
    
    # 构建完整数据包 - 使用与ip-api.com完全一致的字段名称
    data = {
        # 核心字段 (必需)
        'id': analytics_id,
        'user': user_hash,
        'ts': timestamp,
        
        # 基础环境信息 
        'platform': platform_info,
        'ua': f'Python-OmicVerse/{platform.python_version()}',
        'lang': 'en-US',  
        'tz': geo_data.get('timezone', 'UTC'),          # ✅ 向后兼容
        'timezone': geo_data.get('timezone', 'UTC'),    # ✅ 与ip-api.com完全一致
        'ref': '',  # Python环境没有referrer
        
        # 基础地理信息 (必需)
        'country': geo_data.get('country', 'Unknown'),
        
        # 详细地理信息 - 使用ip-api.com的确切字段名称
        'status': geo_data.get('status', 'success'),         # ✅ ip-api状态字段
        'countryCode': geo_data.get('countryCode', 'XX'),    # ✅ 驼峰式，与ip-api一致
        'region': geo_data.get('region', 'Unknown'),         # ✅ 州/省代码 (如 "CA")
        'regionName': geo_data.get('regionName', 'Unknown'), # ✅ 州/省全名 (如 "California")
        'city': geo_data.get('city', 'Unknown'),             # ✅ 城市名
        'zip': geo_data.get('zip', ''),                      # ✅ 邮编 (字段名为zip，不是zipCode)
        'lat': geo_data.get('lat', 0),                       # ✅ 纬度 (数字类型)
        'lon': geo_data.get('lon', 0),                       # ✅ 经度 (数字类型)
        'isp': geo_data.get('isp', 'Unknown'),               # ✅ ISP提供商
        'org': geo_data.get('org', 'Unknown'),               # ✅ 组织名 (简化版)
        'as': geo_data.get('as', 'Unknown'),                 # ✅ AS信息
        'query': geo_data.get('query', 'Unknown'),           # ✅ 查询的IP地址
        
        **kwargs          # 额外参数
    }
    
    # 添加事件类型
    if event_type != 'report_view':
        data['event_type'] = event_type
    
    print(f"📊 发送完整Analytics数据: {analytics_id}")
    print(f"👤 用户Hash: {user_hash}")
    print(f"📍 详细位置: {data['city']}, {data['regionName']} ({data['region']}), {data['country']}")
    print(f"📮 邮编: {data['zip']}")
    print(f"🕐 时区: {data['tz']}")  
    print(f"🌐 语言: {data['lang']}")  
    print(f"🌐 ISP: {data['isp']}")
    print(f"🏢 组织: {data['org']}")
    print(f"📍 精确坐标: ({data['lat']}, {data['lon']})")
    print(f"🔌 AS信息: {data['as']}")
    print(f"⏰ 时间: {timestamp}")
    
    # 发送请求
    analytics_endpoint = os.environ.get('OMICVERSE_ANALYTICS_ENDPOINT', 'http://8.130.139.217/track.gif')
    
    try:
        params = urlencode(data)
        url = f"{analytics_endpoint}?{params}"
        
        print(f"📡 发送到: {analytics_endpoint}")
        print(f"📋 参数数量: {len(data)} 个字段")
        print(f"🔍 核心参数: lang={data['lang']}, tz={data['tz']}, country={data['country']}")
        print(f"🌍 地理详情: city={data['city']}, region={data['region']}, regionName={data['regionName']}")
        print(f"📍 坐标与ISP: lat={data['lat']}, lon={data['lon']}, isp={data['isp']}")
        
        response = requests.get(url, timeout=10)
        
        if response.status_code == 200:
            print("✅ 完整Analytics数据发送成功!")
            print(f"📊 发送了 {len(data)} 个字段，完全匹配ip-api.com结构")
            return True
        else:
            print(f"❌ 发送失败: HTTP {response.status_code}")
            return False
            
    except Exception as e:
        print(f"❌ 发送失败: {str(e)}")
        return False


def test_analytics_connection():
    """
    测试与analytics服务器的连接
    """
    print("🧪 测试Analytics服务器连接...")
    
    # 测试健康检查端点
    health_endpoint = 'http://8.130.139.217/health'
    try:
        response = requests.get(health_endpoint, timeout=5)
        if response.status_code == 200:
            health_data = response.json()
            print(f"✅ 服务器健康: {health_data.get('status')}")
            print(f"📅 服务器时间: {health_data.get('timestamp')}")
            print(f"🔖 服务器版本: {health_data.get('version')}")
        else:
            print(f"⚠️ 健康检查失败: HTTP {response.status_code}")
    except Exception as e:
        print(f"❌ 无法连接到服务器: {e}")
        return False
    
    # 测试debug端点
    debug_endpoint = 'http://8.130.139.217/debug'
    try:
        test_params = {'test': 'connection', 'country': 'TestCountry'}
        response = requests.get(debug_endpoint, params=test_params, timeout=5)
        if response.status_code == 200:
            debug_data = response.json()
            print(f"🔍 Debug测试成功")
            print(f"📊 收到参数: {debug_data.get('url_params', {})}")
        else:
            print(f"⚠️ Debug测试失败: HTTP {response.status_code}")
    except Exception as e:
        print(f"❌ Debug测试失败: {e}")
    
    return True


def send_analytics_silent(analytics_id, event_type='report_view', **kwargs):
    """
    静默版发送函数 - 不输出任何信息
    
    Parameters:
    -----------
    analytics_id : str
        项目或报告的唯一标识符
    event_type : str
        事件类型，默认为'report_view'
    **kwargs : dict
        额外的参数
        
    Returns:
    --------
    bool : 发送是否成功
    
    Note:
    -----
    与send_analytics功能完全相同，但不输出任何调试信息
    """
    
    try:
        # 基础数据收集
        user_hash = generate_user_hash()
        timestamp = datetime.now().isoformat()
        platform_info = platform.system() + " " + platform.release()
        
        # 获取地理位置
        geo_data = get_geolocation()
        
        # 构建数据包
        data = {
            # 核心字段  
            'id': analytics_id,
            'user': user_hash,
            'ts': timestamp,
            
            # 基础环境信息
            'platform': platform_info,
            'ua': f'Python-OmicVerse/{platform.python_version()}',
            'lang': 'en-US',
            'tz': geo_data.get('timezone', 'UTC'),
            'country': geo_data.get('country', 'Unknown'),
            'ref': '',
            
            **kwargs
        }
        
        if event_type != 'report_view':
            data['event_type'] = event_type
        
        # 发送请求
        analytics_endpoint = os.environ.get('OMICVERSE_ANALYTICS_ENDPOINT', 'http://8.130.139.217/track.gif')
        
        params = urlencode(data)
        url = f"{analytics_endpoint}?{params}"
        
        response = requests.get(url, timeout=10)
        return response.status_code == 200
        
    except:
        return False


def send_analytics_full_silent(analytics_id, event_type='report_view', **kwargs):
    """
    静默版完整发送函数 - 不输出任何信息
    
    Parameters:
    -----------
    analytics_id : str
        项目或报告的唯一标识符
    event_type : str
        事件类型，默认为'report_view'
    **kwargs : dict
        额外的参数
        
    Returns:
    --------
    bool : 发送是否成功
    
    Note:
    -----
    与send_analytics_full功能完全相同，但不输出任何调试信息
    """
    
    try:
        # 基础数据收集
        user_hash = generate_user_hash()
        timestamp = datetime.now().isoformat()
        platform_info = platform.system() + " " + platform.release()
        
        # 获取完整地理位置
        geo_data = get_geolocation()
        
        # 构建完整数据包
        data = {
            # 核心字段
            'id': analytics_id,
            'user': user_hash,
            'ts': timestamp,
            
            # 基础环境信息 
            'platform': platform_info,
            'ua': f'Python-OmicVerse/{platform.python_version()}',
            'lang': 'en-US',  
            'tz': geo_data.get('timezone', 'UTC'),
            'timezone': geo_data.get('timezone', 'UTC'),
            'ref': '',
            
            # 基础地理信息
            'country': geo_data.get('country', 'Unknown'),
            
            # 详细地理信息 - ip-api.com兼容
            'status': geo_data.get('status', 'success'),
            'countryCode': geo_data.get('countryCode', 'XX'),
            'region': geo_data.get('region', 'Unknown'),
            'regionName': geo_data.get('regionName', 'Unknown'),
            'city': geo_data.get('city', 'Unknown'),
            'zip': geo_data.get('zip', ''),
            'lat': geo_data.get('lat', 0),
            'lon': geo_data.get('lon', 0),
            'isp': geo_data.get('isp', 'Unknown'),
            'org': geo_data.get('org', 'Unknown'),
            'as': geo_data.get('as', 'Unknown'),
            'query': geo_data.get('query', 'Unknown'),
            
            **kwargs
        }
        
        if event_type != 'report_view':
            data['event_type'] = event_type
        
        # 发送请求
        analytics_endpoint = os.environ.get('OMICVERSE_ANALYTICS_ENDPOINT', 'http://8.130.139.217/track.gif')
        
        params = urlencode(data)
        url = f"{analytics_endpoint}?{params}"
        
        response = requests.get(url, timeout=10)
        return response.status_code == 200
        
    except:
        return False


def send_analytics_simple_silent(analytics_id):
    """
    静默版简化发送函数 - 不输出任何信息
    
    Parameters:
    -----------
    analytics_id : str
        项目标识符
        
    Returns:
    --------
    bool : 发送是否成功
    """
    return send_analytics_silent(analytics_id)


if __name__ == "__main__":
    # 测试用例
    print("🧬 OmicVerse Analytics Sender 测试")
    print("=" * 50)
    
    # 测试连接
    if test_analytics_connection():
        print("\n" + "=" * 50)
        
        # 发送简化版测试数据
        test_id_simple = f"TEST-SIMPLE-{datetime.now().strftime('%Y%m%d-%H%M%S')}"
        print(f"\n📤 测试1: 标准版Analytics (包含基础环境信息)")
        print("-" * 30)
        success1 = send_analytics_simple(test_id_simple)
        
        print(f"\n" + "=" * 50)
        
        # 发送完整版测试数据
        test_id_full = f"TEST-FULL-{datetime.now().strftime('%Y%m%d-%H%M%S')}"
        print(f"\n📤 测试2: 完整版Analytics (包含详细地理信息)")
        print("-" * 30)
        success2 = send_analytics_full(test_id_full, project_name="Test Project", analysis_type="scRNA-seq")
        
        # 总结
        print(f"\n" + "=" * 50)
        print(f"📊 测试结果总结:")
        print(f"✅ 标准版: {'成功' if success1 else '失败'} - ID: {test_id_simple}")
        print(f"✅ 完整版: {'成功' if success2 else '失败'} - ID: {test_id_full}")
        
        if success1 or success2:
            print(f"\n🎉 测试成功! 你可以在dashboard中查看这些记录")
            print(f"📋 Dashboard: http://8.130.139.217/dashboard")
        else:
            print(f"\n❌ 所有测试失败")
    
    print(f"\n📖 使用方法:")
    print(f"# 标准版 (推荐日常使用，包含基础环境信息)")
    print(f"import omicverse.single._analytics_sender as sender")
    print(f"sender.send_analytics_simple('YOUR-PROJECT-ID')")
    print(f"")
    print(f"# 完整版 (包含详细地理信息：城市、ISP、坐标等)")
    print(f"sender.send_analytics_full('YOUR-PROJECT-ID', project_name='My Analysis')")
    print(f"")
    print(f"# 静默版本 (不输出任何调试信息)")
    print(f"success = sender.send_analytics_simple_silent('YOUR-PROJECT-ID')")
    print(f"success = sender.send_analytics_full_silent('YOUR-PROJECT-ID')")
    print(f"# 返回值: True=成功, False=失败") 


from ._flow_expressions import construct_flows_from_cellchat, construct_flows_from_commot, construct_flows_from_cellphonedb, construct_flows_from_liana
from ._flow_preprocessing import determine_informative_variables
from ._gem_construction import construct_gems_using_pyliger, construct_gems_using_nsf, construct_gems_using_nmf
from ._spatial_blocking import construct_spatial_blocks



#!/usr/bin/env python
"""
快速编译工具，用于编译fa2util模块以获得10-100倍的性能提升
"""

import os
import sys
import subprocess
import shutil

def clean_build_files():
    """清理之前的编译文件和缓存"""
    print("清理旧的编译文件...")
    
    # 获取当前脚本所在的目录
    script_dir = os.path.dirname(os.path.abspath(__file__))
    
    # 删除编译产生的文件和目录
    extensions = ['.c', '.cpp', '.html', '.so', '.pyd', '.dll']
    files_to_remove = []
    
    # 查找可能的编译输出文件
    for file in os.listdir(script_dir):
        if file.startswith('fa2util') and any(file.endswith(ext) for ext in extensions):
            files_to_remove.append(os.path.join(script_dir, file))
            
    # 删除build目录
    build_dir = os.path.join(script_dir, 'build')
    if os.path.exists(build_dir):
        files_to_remove.append(build_dir)
    
    # 删除__pycache__目录
    pycache_dir = os.path.join(script_dir, '__pycache__')
    if os.path.exists(pycache_dir):
        files_to_remove.append(pycache_dir)
    
    # 执行删除
    for path in files_to_remove:
        try:
            if os.path.isdir(path):
                shutil.rmtree(path)
                print(f"已删除目录: {path}")
            else:
                os.remove(path)
                print(f"已删除文件: {path}")
        except Exception as e:
            print(f"无法删除 {path}: {str(e)}")
    
    print("清理完成")

def fix_pxd_file():
    """检查并修复fa2util.pxd文件中的已知问题"""
    pxd_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'fa2util.pxd')
    
    if not os.path.exists(pxd_file):
        print(f"警告: 找不到 {pxd_file}")
        return
    
    try:
        with open(pxd_file, 'r') as f:
            content = f.read()
        
        # 修复已知问题: =None 默认值在pxd文件中是不允许的
        if "all_nodes=None" in content:
            print("修复pxd文件中的语法问题...")
            content = content.replace("all_nodes=None", "all_nodes=*")
            with open(pxd_file, 'w') as f:
                f.write(content)
            print("已修复 fa2util.pxd")
    except Exception as e:
        print(f"修复pxd文件时出错: {str(e)}")

def main():
    print("开始编译fa2util模块...")
    
    # 获取当前脚本所在的目录
    script_dir = os.path.dirname(os.path.abspath(__file__))
    
    # 确保我们在正确的目录中
    os.chdir(script_dir)
    
    # 清理旧的编译文件
    clean_build_files()
    
    # 修复pxd文件中的已知问题
    fix_pxd_file()
    
    # 检查是否安装了所需的包
    try:
        import cython
        import numpy
    except ImportError:
        print("缺少必要的依赖，正在安装...")
        subprocess.check_call([sys.executable, "-m", "pip", "install", "cython", "numpy"])
    
    # 执行编译
    try:
        subprocess.check_call([sys.executable, "setup.py", "build_ext", "--inplace"])
        
        # 验证编译结果
        compiled_files = [f for f in os.listdir(script_dir) 
                         if f.startswith('fa2util') and (f.endswith('.so') or f.endswith('.pyd'))]
        
        if compiled_files:
            print("编译成功完成！")
            print(f"生成的文件: {', '.join(compiled_files)}")
            print("现在ForceAtlas2应该能以10-100倍的速度运行。")
            return 0
        else:
            print("警告: 编译似乎成功，但找不到编译后的模块文件。")
            print("ForceAtlas2仍将使用较慢的Python实现。")
            return 1
    except subprocess.CalledProcessError as e:
        print(f"编译失败，错误代码: {e.returncode}")
        print("请确保已安装Cython和必要的编译工具。")
        print("对于Windows用户，您可能需要安装Visual C++ Build Tools。")
        print("对于Mac用户，您可能需要安装XCode命令行工具。")
        print("对于Linux用户，请确保已安装gcc和python-dev包。")
        print("\n如果您在服务器上运行，可能需要使用以下命令安装依赖:")
        print("Debian/Ubuntu: sudo apt-get install python3-dev build-essential")
        print("CentOS/RHEL: sudo yum install python3-devel gcc")
        print("\n您也可以尝试使用GPU版本，它不需要编译:")
        print("fle(..., use_gpu=True, compile_cython=False)")
        return 1
    
if __name__ == "__main__":
    sys.exit(main()) 


"""
Like image_sample.py, but use a noisy image classifier to guide the sampling
process towards more realistic images.
"""

import argparse

import numpy as np
import torch as th
import torch.distributed as dist
import torch.nn.functional as F

from .guided_diffusion import dist_util, logger
from .guided_diffusion.script_util import (   
    NUM_CLASSES,
    model_and_diffusion_defaults,
    classifier_and_diffusion_defaults,
    create_model_and_diffusion,
    create_classifier,
    add_dict_to_argparser,
    args_to_dict,
)
import scanpy as sc
import torch
from VAE.VAE_model import VAE

def load_VAE(ae_dir, num_gene):
    autoencoder = VAE(
        num_genes=num_gene,
        device='cuda',
        seed=0,
        hidden_dim=128,
        decoder_activation='ReLU',
    )
    autoencoder.load_state_dict(torch.load(ae_dir))
    return autoencoder

def save_data(all_cells, traj, data_dir):
    cell_gen = all_cells
    np.savez(data_dir, cell_gen=cell_gen)
    return

def main(cell_type=[0], multi=False, inter=False, weight=[10,10]):
    args = create_argparser(cell_type, weight).parse_args()

    dist_util.setup_dist()
    logger.configure()

    logger.log("creating model and diffusion...")
    model, diffusion = create_model_and_diffusion(
        **args_to_dict(args, model_and_diffusion_defaults().keys())
    )
    model.load_state_dict(
        dist_util.load_state_dict(args.model_path, map_location="cpu")
    )
    model.to(dist_util.dev())
    model.eval()

    logger.log("loading classifier...")
    if multi:
        args.num_class = args.num_class1 # how many classes in this condition
        classifier1 = create_classifier(**args_to_dict(args, (['num_class']+list(classifier_and_diffusion_defaults().keys()))[:3]))
        classifier1.load_state_dict(
            dist_util.load_state_dict(args.classifier_path1, map_location="cpu")
        )
        classifier1.to(dist_util.dev())
        classifier1.eval()

        args.num_class = args.num_class2 # how many classes in this condition
        classifier2 = create_classifier(**args_to_dict(args, (['num_class']+list(classifier_and_diffusion_defaults().keys()))[:3]))
        classifier2.load_state_dict(
            dist_util.load_state_dict(args.classifier_path2, map_location="cpu")
        )
        classifier2.to(dist_util.dev())
        classifier2.eval()

    else:
        classifier = create_classifier(**args_to_dict(args, (['num_class']+list(classifier_and_diffusion_defaults().keys()))[:3]))
        classifier.load_state_dict(
            dist_util.load_state_dict(args.classifier_path, map_location="cpu")
        )
        classifier.to(dist_util.dev())
        classifier.eval()

    '''
    control function for Gradient Interpolation Strategy
    '''
    def cond_fn_inter(x, t, y=None, init=None, diffusion=None):
        assert y is not None
        y1 = y[:,0]
        y2 = y[:,1]
        # xt = diffusion.q_sample(th.tensor(init,device=dist_util.dev()),t*th.ones(init.shape[0],device=dist_util.dev(),dtype=torch.long),)
        with th.enable_grad():
            x_in = x.detach().requires_grad_(True)
            logits = classifier(x_in, t)
            log_probs = F.log_softmax(logits, dim=-1)
            selected1 = log_probs[range(len(logits)), y1.view(-1)]
            selected2 = log_probs[range(len(logits)), y2.view(-1)]
            
            grad1 = th.autograd.grad(selected1.sum(), x_in, retain_graph=True)[0] * args.classifier_scale1
            grad2 = th.autograd.grad(selected2.sum(), x_in, retain_graph=True)[0] * args.classifier_scale2

            # l2_loss = ((x_in-xt)**2).mean()
            # grad3 = th.autograd.grad(-l2_loss, x_in, retain_graph=True)[0] * 100

            return grad1+grad2#+grad3

    '''
    control function for multi-conditional generation
    Two conditional generation here
    '''
    def cond_fn_multi(x, t, y=None):
        assert y is not None
        y1 = y[:,0]
        y2 = y[:,1]
        with th.enable_grad():
            x_in = x.detach().requires_grad_(True)
            logits1 = classifier1(x_in, t)
            log_probs1 = F.log_softmax(logits1, dim=-1)
            selected1 = log_probs1[range(len(logits1)), y1.view(-1)]

            logits2 = classifier2(x_in, t)
            log_probs2 = F.log_softmax(logits2, dim=-1)
            selected2 = log_probs2[range(len(logits2)), y2.view(-1)]
            
            grad1 = th.autograd.grad(selected1.sum(), x_in, retain_graph=True)[0] * args.classifier_scale1
            grad2 = th.autograd.grad(selected2.sum(), x_in, retain_graph=True)[0] * args.classifier_scale2
            
            return grad1+grad2

    '''
    control function for one conditional generation
    '''
    def cond_fn_ori(x, t, y=None):
        assert y is not None
        with th.enable_grad():
            x_in = x.detach().requires_grad_(True)
            logits = classifier(x_in, t)
            log_probs = F.log_softmax(logits, dim=-1)
            selected = log_probs[range(len(logits)), y.view(-1)]
            grad = th.autograd.grad(selected.sum(), x_in, retain_graph=True)[0] * args.classifier_scale
            return grad
        
    def model_fn(x, t, y=None, init=None, diffusion=None):
        assert y is not None
        if args.class_cond:
            return model(x, t, y if args.class_cond else None)
        else:
            return model(x, t)
        
    if inter:
        # input real cell expression data as initial noise
        ori_adata = sc.read_h5ad(args.init_cell_path)
        sc.pp.normalize_total(ori_adata, target_sum=1e4)
        sc.pp.log1p(ori_adata)

    logger.log("sampling...")
    all_cell = []
    sample_num = 0
    while sample_num < args.num_samples:
        model_kwargs = {}

        if not multi and not inter:
            classes = (cell_type[0])*th.ones((args.batch_size,), device=dist_util.dev(), dtype=th.long)

        if multi:
            classes1 = (cell_type[0])*th.ones((args.batch_size,), device=dist_util.dev(), dtype=th.long)
            classes2 = (cell_type[1])*th.ones((args.batch_size,), device=dist_util.dev(), dtype=th.long)
            # classes3 = ... if more conditions
            classes = th.stack((classes1,classes2), dim=1)

        if inter:
            classes1 = (cell_type[0])*th.ones((args.batch_size,), device=dist_util.dev(), dtype=th.long)
            classes2 = (cell_type[1])*th.ones((args.batch_size,), device=dist_util.dev(), dtype=th.long)
            classes = th.stack((classes1,classes2), dim=1)

        model_kwargs["y"] = classes
        sample_fn = (
            diffusion.p_sample_loop if not args.use_ddim else diffusion.ddim_sample_loop
        )

        if inter:
            celltype = ori_adata.obs['period'].cat.categories.tolist()[cell_type[0]]
            adata = ori_adata[ori_adata.obs['period']==celltype].copy()

            start_x = adata.X
            autoencoder = load_VAE(args.ae_dir, args.num_gene)
            start_x = autoencoder(torch.tensor(start_x,device=dist_util.dev()),return_latent=True).detach().cpu().numpy()

            n, m = start_x.shape  
            if n >= args.batch_size:  
                start_x = start_x[:args.batch_size, :]  
            else:  
                repeat_times = args.batch_size // n  
                remainder = args.batch_size % n  
                start_x = np.concatenate([start_x] * repeat_times + [start_x[:remainder, :]], axis=0)  
            
            noise = diffusion.q_sample(th.tensor(start_x,device=dist_util.dev()),args.init_time*th.ones(start_x.shape[0],device=dist_util.dev(),dtype=torch.long),)
            model_kwargs["init"] = start_x
            model_kwargs["diffusion"] = diffusion

        if multi:
            sample, traj = sample_fn(
                model_fn,
                (args.batch_size, args.input_dim),
                clip_denoised=args.clip_denoised,
                model_kwargs=model_kwargs,
                cond_fn=cond_fn_multi,
                device=dist_util.dev(),
                noise = None,
                start_time=diffusion.betas.shape[0],
                start_guide_steps=args.start_guide_steps,
            )
        elif inter:
            sample, traj = sample_fn(
                model_fn,
                (args.batch_size, args.input_dim),
                clip_denoised=args.clip_denoised,
                model_kwargs=model_kwargs,
                cond_fn=cond_fn_inter,
                device=dist_util.dev(),
                noise = noise,
                start_time=diffusion.betas.shape[0],
                start_guide_steps=args.start_guide_steps,
            )
        else:
            sample, traj = sample_fn(
                model_fn,
                (args.batch_size, args.input_dim),
                clip_denoised=args.clip_denoised,
                model_kwargs=model_kwargs,
                cond_fn=cond_fn_ori,
                device=dist_util.dev(),
                noise = None,
            )

        gathered_samples = [th.zeros_like(sample) for _ in range(dist.get_world_size())]
        dist.all_gather(gathered_samples, sample)  # gather not supported with NCCL
        if args.filter:
            for sample in gathered_samples:
                if multi:
                    logits1 = classifier1(sample, torch.zeros((sample.shape[0]), device=sample.device))
                    logits2 = classifier2(sample, torch.zeros((sample.shape[0]), device=sample.device))
                    prob1 = F.softmax(logits1, dim=-1)
                    prob2 = F.softmax(logits2, dim=-1)
                    type1 = torch.argmax(prob1, 1)
                    type2 = torch.argmax(prob2, 1)
                    select_index = ((type1 == cell_type[0]) & (type2 == cell_type[1]))
                    all_cell.extend([sample[select_index].cpu().numpy()])
                    sample_num += select_index.sum().item()
                elif inter:
                    logits = classifier(sample, torch.zeros((sample.shape[0]), device=sample.device))
                    prob = F.softmax(logits, dim=-1)
                    left = (prob[:,cell_type[0]] > weight[0]/10-0.15) & (prob[:,cell_type[0]] < weight[0]/10+0.15)
                    right = (prob[:,cell_type[1]] > weight[1]/10-0.15) & (prob[:,cell_type[1]] < weight[1]/10+0.15)
                    select_index = left & right
                    all_cell.extend([sample[select_index].cpu().numpy()])
                    sample_num += select_index.sum().item()
                else:
                    logits = classifier(sample, torch.zeros((sample.shape[0]), device=sample.device))
                    prob = F.softmax(logits, dim=-1)
                    type = torch.argmax(prob, 1)
                    select_index = (type == cell_type[0])
                    all_cell.extend([sample[select_index].cpu().numpy()])
                    sample_num += select_index.sum().item()
            logger.log(f"created {sample_num} samples")
        else:
            all_cell.extend([sample.cpu().numpy() for sample in gathered_samples])
            sample_num = len(all_cell) * args.batch_size
            logger.log(f"created {len(all_cell) * args.batch_size} samples")

    arr = np.concatenate(all_cell, axis=0)
    save_data(arr, traj, args.sample_dir+str(cell_type[0]))

    dist.barrier()
    logger.log("sampling complete")


def create_argparser(celltype=[0], weight=[10,10]):
    defaults = dict(
        clip_denoised=True,
        num_samples=9000,
        batch_size=3000,
        use_ddim=False,
        class_cond=False, 

        model_path="output/diffusion_checkpoint/muris_diffusion/model000000.pt", 

        # ***if commen conditional generation & gradiante interpolation, use this path***
        classifier_path="output/classifier_checkpoint/classifier_muris/model000100.pt",
        # ***if multi-conditional, use this path. replace this to your own classifiers***
        classifier_path1="output/classifier_checkpoint/classifier_muris_ood_type/model200000.pt",
        classifier_path2="output/classifier_checkpoint/classifier_muris_ood_organ/model200000.pt",
        num_class1 = 2,  # set this to the number of classes in your own dataset. this is the first condition (for example cell organ).
        num_class2 = 2,  # this is the second condition (for example cell type).

        # ***if commen conditional generation, use this scale***
        classifier_scale=2,
        # ***in multi-conditional, use this scale. scale1 and scale2 are the weights of two classifiers***
        # ***in Gradient Interpolation, use this scale, too. scale1 and scale2 are the weights of two gradients***
        classifier_scale1=weight[0]*2/10,
        classifier_scale2=weight[1]*2/10,

        # ***if gradient interpolation, replace these base on your own situation***
        ae_dir='output/Autoencoder_checkpoint/WOT/model_seed=0_step=150000.pt', 
        num_gene=19423,
        init_time = 600,    # initial noised state if interpolation
        init_cell_path = 'data/WOT/filted_data.h5ad',   #input initial noised cell state

        sample_dir=f"output/simulated_samples/muris",
        start_guide_steps = 500,     # the time to use classifier guidance
        filter = False,   # filter the simulated cells that are classified into other condition, might take long time

    )
    defaults.update(model_and_diffusion_defaults())
    defaults.update(classifier_and_diffusion_defaults())
    defaults['num_class']=12
    parser = argparse.ArgumentParser()
    add_dict_to_argparser(parser, defaults)
    return parser


if __name__ == "__main__":
    # for conditional generation
    # main(cell_type=[2])
    for type in range(12):
        main(cell_type=[type])

    # ***for multi-condition, run***
    # muris ood
    # for i in [0,1]:
    #     for j in [0,1]:
    #         main(cell_type=[i,j],multi=True)

    # ***for Gradient Interpolation, run***
    # for i in range(0,11):
    #     main(cell_type=[6,7], inter=True, weight=[10-i,i])
    # for i in range(18):
    #     main(cell_type=[i,i+1], inter=True, weight=[5,5])


r"""
Training functions with different strategy
"""
from math import ceil
from typing import List, Optional, Union

import torch
import torch.nn.functional as F

from .loss import feature_reconstruct_loss


def train_GAN(wdiscriminator:torch.nn.Module,
                optimizer_d:torch.optim.Optimizer,
                embds:List[torch.Tensor],
                batch_d_per_iter:Optional[int]=5,
                anchor_scale:Optional[float]=0.8
    )->torch.Tensor:
    r"""
    GAN training strategy
    
    Parameters
    ----------
    wdiscriminator
        WGAN
    optimizer_d
        WGAN parameters
    embds
        list of LGCN embd
    batch_d_per_iter
        WGAN train iter numbers
    anchor_scale
        ratio of anchor cells
    """
    embd0, embd1 = embds
    
    wdiscriminator.train()
    anchor_size = ceil(embd1.size(0)*anchor_scale)

    for j in range(batch_d_per_iter):
        w0 = wdiscriminator(embd0)
        w1 = wdiscriminator(embd1)
        anchor1 = w1.view(-1).argsort(descending=True)[: anchor_size]
        anchor0 = w0.view(-1).argsort(descending=False)[: anchor_size]
        embd0_anchor = embd0[anchor0, :].clone().detach()
        embd1_anchor = embd1[anchor1, :].clone().detach()
        optimizer_d.zero_grad()
        loss = -torch.mean(wdiscriminator(embd0_anchor)) + torch.mean(wdiscriminator(embd1_anchor))
        loss.backward()
        optimizer_d.step()
        for p in wdiscriminator.parameters():
            p.data.clamp_(-0.1, 0.1)
    w0 = wdiscriminator(embd0)
    w1 = wdiscriminator(embd1)
    anchor1 = w1.view(-1).argsort(descending=True)[: anchor_size]
    anchor0 = w0.view(-1).argsort(descending=False)[: anchor_size]
    embd0_anchor = embd0[anchor0, :]
    embd1_anchor = embd1[anchor1, :]
    loss = -torch.mean(wdiscriminator(embd1_anchor))
    return loss


def train_reconstruct(recon_models:torch.nn.Module,
                        optimizer_recons,
                        embds:List[torch.Tensor],
                        features:List[torch.Tensor],
                        batch_r_per_iter:Optional[int]=10
    )->torch.Tensor:
    r"""
    Data reconstruction network training strategy
    
    Parameters
    ----------
    recon_models
        list of reconstruction model
    optimizer_recons
        list of reconstruction optimizer
    embds
        list of LGCN embd
    features
        list of rae node features
    batch_d_per_iter
        WGAN train iter numbers
    """
    recon_model0, recon_model1 = recon_models
    optimizer_recon0, optimizer_recon1 = optimizer_recons
    embd0, embd1 = embds
    
    recon_model0.train()
    recon_model1.train()
    embd0_copy = embd0.clone().detach()
    embd1_copy = embd1.clone().detach()    
    for t in range(batch_r_per_iter):
        optimizer_recon0.zero_grad()
        loss = feature_reconstruct_loss(embd0_copy, features[0], recon_model0)
        loss.backward()
        optimizer_recon0.step()
    for t in range(batch_r_per_iter):
        optimizer_recon1.zero_grad()
        loss = feature_reconstruct_loss(embd1_copy, features[1], recon_model1)
        loss.backward()
        optimizer_recon1.step()
    loss = 0.5 * feature_reconstruct_loss(embd0, features[0], recon_model0) + 0.5 * feature_reconstruct_loss(embd1, features[1], recon_model1)

    return loss
    
        
def check_align(embds:List[torch.Tensor],
                ground_truth:torch.Tensor,
                k:Optional[int]=[5,10],
                mode:Optional[str]='cosine'
    )->List[float]:
    r"""
    Check embedding correspondence in given distance (default cosine similarity) under ground truth
    
    Parameters
    -----------
    embds
        List of graph features, each element is (node_num, feature_dim)
    ground_truth
        mapping ground_truth (2, node_num)
    k
        list of top k (only support 2 elements yet)
    mode
        distance quota
    """
    embd0, embd1 = embds
    assert k[1] > k[0]
    g_map = {}
    for i in range(ground_truth.size(1)):
        g_map[ground_truth[1, i].item()] = ground_truth[0, i].item()
    g_list = list(g_map.keys())
    
    cossim = torch.zeros(embd1.size(0), embd0.size(0))
    for i in range(embd1.size(0)):
        cossim[i] = F.cosine_similarity(embd0, embd1[i:i+1].expand(embd0.size(0), embd1.size(1)), dim=-1).view(-1)
    
    ind = cossim.argsort(dim=1, descending=True)[:, :k[1]]
    a1 = 0
    ak0 = 0
    ak1 = 0
    for i, node in enumerate(g_list):
        if ind[node, 0].item() == g_map[node]:
            a1 += 1
            ak0 += 1
            ak1 += 1
        else:
            for j in range(1, k[0]):
                if ind[node, j].item() == g_map[node]:
                    ak0 += 1
                    ak1 += 1
                    break
                else:
                    for l in range(k[0], k[1]):
                        if ind[node, l].item() == g_map[node]:
                            ak1 += 1
                        break

    a1 /= len(g_list)
    ak0 /= len(g_list)
    ak1 /= len(g_list)
    print(f'H@1:{a1*100}; H@{k[0]}:{ak0*100}; H@{k[1]}:{ak1*100}')
    return a1, ak0, ak1


import numpy as np
import scipy as s
import scipy.stats as stats
from .basic_distributions import Distribution

from ..utils import *

class Poisson(Distribution):
    """
    Class to define Poisson distributions.

    Equations:
    p(x|theta) = theta**x * exp(-theta) * 1/theta!
    log p(x|a,b) = x*theta - theta - log(x!)
    E[x] = theta
    var[x] = theta
    """

    def __init__(self, dim, theta, E=None):
        Distribution.__init__(self, dim)

        # Initialise parameters
        theta = np.ones(dim) * theta
        self.params = { 'theta':theta }

        # Initialise expectations
        if E is None:
            self.updateExpectations()
        else:
            self.expectations = { 'E':np.ones(dim)*E }

        # Check that dimensionalities match
        self.CheckDimensionalities()

    def updateExpectations(self):
        E = self.params['theta']
        self.expectations = { 'E':E }

    def density(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        assert x.dtype == int, "x has to be an integer array"
        theta = self.params['theta'].flatten()
        x = x.flatten()
        # return np.prod (stats.poisson.pmf(x,theta) )
        return np.prod( np.divide(theta**x * np.exp(-theta),s.misc.factorial(x)) )

    def loglik(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        assert x.dtype == int, "x has to be an integer array"
        theta = self.params['theta'].flatten()
        x = x.flatten()
        # return np.log( np.prod (stats.poisson.pmf(x,theta) ))
        return np.sum( x*np.log(theta) - theta - np.log(s.misc.factorial(x)) )



r"""
Visualization modules of SLAT 
"""
from typing import Optional
import matplotlib.pyplot as plt
from matplotlib import colors

from .multi_dataset import *


def hist(data,
         bins: Optional[int]=100,
         cut: Optional[float]=0.9,
         cut_height: Optional[int]=200
    )->None:
    r"""
    Histogram of distribution
    
    Parameters:
    -----------
    data
        array or list
    bins
        number of bins
    cut
        cutoff line location
    cut_height
        cutoff line height
    """
    plt.hist(data, bins=bins, facecolor="blue", edgecolor="blue", alpha=0.7)
    plt.vlines(cut, 0, cut_height, color="red", linestyles='dotted', label=str(cut))
    plt.xlabel(r'Similarity Score')
    plt.ylabel(r'Cells')
    plt.show()
    

def make_ground(x_shape,y_shape,color='white'):
    r"""
    Make the background of ImageContainer class
    """
    rgb = colors.to_rgb(color)
    ground = np.array(np.ones([x_shape,y_shape,3]) * rgb)
    return ground



from typing import List, Tuple, Optional
import numpy as np
import scanpy as sc

import pandas as pd

def subset_for_flow_type(adata: sc.AnnData,
                         var_type: str = 'all',
                         flowsig_expr_key: str = 'X_flow',
                         flowsig_network_key: str = 'flowsig_network'):
    
    var_types = ['all', 'inflow', 'module', 'outflow']

    if var_type not in var_types:
        ValueError("Need to specify var_type as one of the following: %s"  % var_types)

    X_flow = adata.obsm[flowsig_expr_key]
    adata_subset = sc.AnnData(X=X_flow)
    adata_subset.obs = adata.obs
    adata_subset.var = adata.uns[flowsig_network_key]['flow_var_info']

    if var_type != 'all':

        adata_subset = adata_subset[:, adata_subset.var['Type'] == var_type]

    return adata_subset


def filter_flow_vars(adata: sc.AnnData,
                    vars_subset: List[str],
                    flowsig_expr_key: str = 'X_flow',
                    flowsig_network_key: str = 'flowsig_network'):
    
    X_flow_orig = adata.obsm[flowsig_expr_key]
    flow_var_info_orig = adata.uns[flowsig_network_key]['flow_var_info']
    flowsig_vars_orig = flow_var_info_orig.index.tolist()
    
    # We define the list in this weird way to preserve all the var types etc
    subset_indices = [flowsig_vars_orig.index(flow_var) for flow_var in vars_subset]

    X_flow = X_flow_orig[:, subset_indices]

    # Subset the flowsig network info as well
    flow_var_info = flow_var_info_orig[flow_var_info_orig.index.isin(vars_subset)]

    # Store the new FlowSig variable information
    flowsig_info = {'flow_var_info': flow_var_info}

    adata.obsm[flowsig_expr_key + '_orig'] = X_flow_orig
    adata.obsm[flowsig_expr_key] = X_flow
    adata.uns[flowsig_network_key + '_orig'] = flow_var_info_orig
    adata.uns[flowsig_network_key] = flowsig_info


def determine_differentially_flowing_vars(adata: sc.AnnData,
                                        condition_key: str,
                                        control_key: str,
                                        flowsig_expr_key: str = 'X_flow',
                                        flowsig_network_key: str = 'flowsig_network',
                                        qval_threshold: float = 0.05,
                                        logfc_threshold: float = 0.5):
    
    # Construct AnnData for flow expression
    perturbed_conditions = [cond for cond in adata.obs[condition_key].unique().tolist() if cond != control_key]
    flow_var_info = adata.uns[flowsig_network_key]['flow_var_info']

    # Construct inflow and outflow adata objects
    adata_inflow = subset_for_flow_type(adata,
                                        var_type = 'inflow',
                                        flowsig_expr_key = flowsig_expr_key,
                                        flowsig_network_key = flowsig_network_key)

    
    adata_outflow = subset_for_flow_type(adata,
                                        var_type = 'outflow',
                                        flowsig_expr_key = flowsig_expr_key,
                                        flowsig_network_key = flowsig_network_key)

    # Calculate differentially inflowing vars
    adata_inflow.uns['log1p'] = {'base': None} # Just in case
    sc.tl.rank_genes_groups(adata_inflow, key_added=condition_key, groupby=condition_key, method='wilcoxon')

    # Determine the differentially flowing vars
    diff_inflow_vars = []

    lowqval_des_inflow = {}
    for cond in perturbed_conditions:

        # Get the DEs with respect to this contrast
        result = sc.get.rank_genes_groups_df(adata_inflow, group=cond, key=condition_key).copy()
        result["-logQ"] = -np.log(result["pvals"].astype("float"))
        lowqval_de = result.loc[(np.abs(result["logfoldchanges"]) > logfc_threshold)&(result["pvals_adj"] < qval_threshold)]

        lowqval_des_inflow[cond] = lowqval_de['names'].tolist()
        
    diff_inflow_vars = list(set.union(*map(set, [lowqval_des_inflow[cond] for cond in lowqval_des_inflow])))
    
    # Calculate differentially inflowing vars
    adata_outflow.uns['log1p'] = {'base':None} # Just in case
    sc.tl.rank_genes_groups(adata_outflow, key_added=condition_key, groupby=condition_key, method='wilcoxon')

    # Determine the differentially flowing vars
    diff_outflow_vars = []

    lowqval_des_outflow = {}
    for cond in perturbed_conditions:

        # Get the DEs with respect to this contrast
        result = sc.get.rank_genes_groups_df(adata_outflow, group=cond, key=condition_key).copy()
        result["-logQ"] = -np.log(result["pvals"].astype("float"))
        lowqval_de = result.loc[(np.abs(result["logfoldchanges"]) > logfc_threshold)&(abs(result["pvals_adj"]) < qval_threshold)]

        lowqval_des_outflow[cond] = lowqval_de['names'].tolist()
        
    diff_outflow_vars = list(set.union(*map(set, [lowqval_des_outflow[cond] for cond in lowqval_des_outflow])))

    # We don't change GEM vars because from experience, they typically incorporate condition-specific changes as is
    gem_vars = flow_var_info[flow_var_info['Type'] == 'module'].index.tolist()
    vars_to_subset = diff_inflow_vars + diff_outflow_vars + gem_vars

    filter_flow_vars(adata,
                    vars_to_subset,
                    flowsig_expr_key,
                    flowsig_network_key)
    
def determine_spatially_flowing_vars(adata: sc.AnnData,
                                    flowsig_expr_key: str = 'X_flow',
                                    flowsig_network_key: str = 'flowsig_network',
                                    moran_threshold: float = 0.1,
                                    coord_type: str = 'grid',
                                    n_neighbours: int = 6,
                                    library_key: str = None,
                                    n_perms: int = None,
                                    n_jobs: int = None):
    import squidpy as sq
    # Get the flow info
    flow_var_info = adata.uns[flowsig_network_key]['flow_var_info']
    
    # Construct inflow and outflow adata objects
    adata_inflow = subset_for_flow_type(adata,
                                        var_type = 'inflow',
                                        flowsig_expr_key = flowsig_expr_key,
                                        flowsig_network_key = flowsig_network_key)

    adata_outflow = subset_for_flow_type(adata,
                                        var_type = 'outflow',
                                        flowsig_expr_key = flowsig_expr_key,
                                        flowsig_network_key = flowsig_network_key)
    
    if 'spatial' not in adata.obsm:
        ValueError("Need to specify spatial coordinates in adata.obsm['spatial'].")
    else:
        adata_outflow.obsm['spatial'] = adata.obsm['spatial']
        adata_inflow.obsm['spatial'] = adata.obsm['spatial']

        # Can't have spatial connectivities without spatial coordinates, lol
        if 'spatial_connectivities' not in adata.obsp:

            coord_types = ['grid', 'generic']
            
            if coord_type not in coord_types:
                ValueError("Please specify coord_type to be one of %s" % coord_types)

            sq.gr.spatial_neighbors(adata_outflow, coord_type=coord_type, n_neighs=n_neighbours, library_key=library_key)
            sq.gr.spatial_neighbors(adata_inflow, coord_type=coord_type, n_neighs=n_neighbours, library_key=library_key)

            sq.gr.spatial_autocorr(adata_outflow, genes=adata_outflow.var_names.tolist(), n_perms=n_perms, n_jobs=n_jobs)
            sq.gr.spatial_autocorr(adata_inflow, genes=adata_inflow.var_names.tolist(), n_perms=n_perms, n_jobs=n_jobs)

            # Filter genes based on moran_threshold
            svg_outflows = adata_outflow.uns['moranI'][adata_outflow.uns['moranI']['I'] > moran_threshold].index.tolist()
            svg_inflows = adata_inflow.uns['moranI'][adata_inflow.uns['moranI']['I'] > moran_threshold].index.tolist()
            gem_vars = flow_var_info[flow_var_info['Type'] == 'module'].index.tolist()

            spatially_flowing_vars = svg_outflows + svg_inflows + gem_vars

            # Re-adjust the flow variables
            filter_flow_vars(adata,
                             spatially_flowing_vars,
                             flowsig_expr_key,
                             flowsig_network_key)

def determine_informative_variables(adata: sc.AnnData,  
                                    flowsig_expr_key: str = 'X_flow',
                                    flowsig_network_key: str = 'flowsig_network',
                                    spatial: bool = False,
                                    condition_key: str = None,
                                    control_key: str = None,
                                    moran_threshold: float = 0.1,
                                    qval_threshold: float = 0.05,
                                    logfc_threshold: float = 0.5,
                                    coord_type: str = 'grid',
                                    n_neighbours: int = 6,
                                    library_key: str = None):
    

    if spatial: # We calculate the spatial autocorrelation (using Moran's I) and cut off genes below a defined threshold

        determine_spatially_flowing_vars(adata,
                                         flowsig_expr_key=flowsig_expr_key,
                                         flowsig_network_key=flowsig_network_key,
                                         moran_threshold=moran_threshold,
                                         coord_type=coord_type,
                                         n_neighbours=n_neighbours,
                                         library_key=library_key)
        
    else:

        determine_differentially_flowing_vars(adata,
                                              condition_key=condition_key,
                                              control_key=control_key,
                                              flowsig_expr_key=flowsig_expr_key,
                                              flowsig_network_key=flowsig_network_key,
                                              qval_threshold=qval_threshold,
                                              logfc_threshold=logfc_threshold)



r"""
Loss function
"""
import torch


def feature_reconstruct_loss(embd:torch.Tensor,
                             x:torch.Tensor,
                             recon_model:torch.nn.Module
    ) -> torch.Tensor:
    r"""
    Reconstruction loss (MSE)
    
    Parameters
    ----------
    embd
        embd of a cell
    x
        input 
    recon_model
        reconstruction model
    """
    recon_x = recon_model(embd)
    return torch.norm(recon_x - x, dim=1, p=2).mean()